# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import earthkit.data
import numpy as np
from cartopy.util import add_cyclic_point

from earthkit.maps import styles
from earthkit.maps.schemas import schema

_NO_EARTHKIT_REGRID = False
try:
    import earthkit.regrid
except ImportError:
    _NO_EARTHKIT_REGRID = True


class Input:
    """
    Class for managing input data for charts.
    """

    def __init__(
        self, data, x=None, y=None, transform=None, style=None, domain=None, **kwargs
    ):
        self._data = data
        self._transform = transform

        self._converted_values = None

        self._style = style
        self._kwargs = kwargs

        self.x = x
        self.y = y
        self._values = None
        self.extract(domain)

    @property
    def data(self):
        if not isinstance(self._data, (earthkit.data.core.Base, list, np.ndarray)):
            self._data = earthkit.data.from_object(self._data)
        if isinstance(self._data, earthkit.data.core.Base) and hasattr(
            self._data, "__len__"
        ):
            try:
                self._data = self._data[0]
            except (ValueError, TypeError, AttributeError):
                pass
        return self._data

    @property
    def source_units(self):
        try:
            units = self._data.metadata("units", default=None)
        except (NotImplementedError, AttributeError):
            units = None
        return units

    @property
    def short_name(self):
        try:
            short_name = self._data.metadata("short_name", default="")
        except (NotImplementedError, AttributeError):
            short_name = ""
        return short_name

    @property
    def transform(self):
        if self._transform is None:
            try:
                self._transform = self._data.projection().to_cartopy_crs()
            except AttributeError:
                self._transform = ccrs.PlateCarree()
        return self._transform

    @property
    def style(self):
        if self._style is None:
            style_units = None
            if not schema.use_preferred_styles:
                style_units = self._kwargs.pop("units", None) or self.source_units
            self._style = styles.auto.guess_style(self.data, units=style_units)
        return self._style

    @property
    def values(self):
        return self.style.convert_units(
            self._values,
            self.source_units,
            self.short_name,
        )

    @property
    def gridspec(self):
        if self.data.metadata("gridType", default="") == "reduced_gg":
            n = self.data.metadata("N", default=None)
            if n is not None:
                if self.data.metadata("isOctahedral", default=0):
                    g = f"O{n}"
                else:
                    g = f"N{n}"
            return {"grid": g}

    def extract(self, domain=None):
        if self.x is None and self.y is None:

            if self.gridspec is not None:
                if _NO_EARTHKIT_REGRID:
                    raise ImportError(
                        f"earthkit-regrid is required for plotting data on a"
                        f"'{self.gridspec['grid']}' grid"
                    )
                points = get_points(schema.interpolate_target_resolution)
                self._values = earthkit.regrid.interpolate(
                    self.data.values,
                    self.gridspec,
                    {
                        "grid": [
                            schema.interpolate_target_resolution,
                            schema.interpolate_target_resolution,
                        ],
                    },
                )
                if domain is not None:
                    self._values, points = domain.latlon_bbox(self._values, points)
            else:
                if domain is None:
                    self._values = self.data.to_numpy(flatten=False)
                    points = self.data.to_points(flatten=False)
                else:
                    self._values, points = domain.bbox(self.data)
            self.y = points["y"]
            self.x = points["x"]

        elif self._values is None:
            try:
                self._values = self.data.values()
            except AttributeError:
                self._values = self.data

        if np.all(self.x[:-1] == self.x[1:]):
            self.y = self.y[:, 0]
            self._values, self.x = add_cyclic_point(self._values, coord=self.x[0])
            self.x, self.y = np.meshgrid(self.x, self.y)


class Vector(Input):
    def __init__(self, u, v, *args, **kwargs):
        self._u_data = u
        self._v_data = v
        self.u = None
        self.v = None
        super().__init__(u, *args, **kwargs)

    def extract(self, domain):
        self._data = self._u_data
        super().extract(domain)
        self.u = self._values
        self.x, self.y = None, None

        self._data = self._v_data
        super().extract(domain)
        self.v = self._values

        self._values = [self.u, self.v]


def get_points(dx):
    import numpy as np

    lat_v = np.linspace(90, -90, int(180 / dx) + 1)
    lon_v = np.linspace(0, 360 - dx, int(360 / dx))
    lon, lat = np.meshgrid(lon_v, lat_v)
    return {"x": lon, "y": lat}


def sanitise(data):
    if not isinstance(data, (earthkit.data.core.Base, list, np.ndarray)):
        data = earthkit.data.from_object(data)
    if not isinstance(data, earthkit.data.core.Base) or not hasattr(data, "__len__"):
        data = [data]
    return data[0]



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import cartopy.crs as ccrs
import cartopy.feature as cfeature
import cartopy.io.shapereader as shpreader
import earthkit.data
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from earthkit.maps import domains
from earthkit.maps.domains import natural_earth
from earthkit.maps.layers import metadata
from earthkit.maps.layers.layers import Layer, LayerFormatter
from earthkit.maps.schemas import schema
from earthkit.maps.styles.auto import suggest_style


class SubplotFormatter(metadata.BaseFormatter):
    def __init__(self, subplot, unique=True):
        self.subplot = subplot
        self.unique = unique
        self._layer_index = None

    def convert_field(self, value, conversion):
        f = super().convert_field
        if isinstance(value, list):
            if isinstance(conversion, str) and conversion.isnumeric():
                return str(value[int(conversion)])
            return [f(v, conversion) for v in value]
        else:
            return f(value, conversion)

    def format_key(self, key):
        values = [
            LayerFormatter(layer).format_key(key) for layer in self.subplot.layers
        ]
        return values

    def format_field(self, value, format_spec):
        f = super().format_field
        if isinstance(value, list):
            values = [f(v, format_spec) for v in value]
            if self._layer_index is not None:
                value = values[self._layer_index]
                self._layer_index = None
            else:
                if self.unique:
                    values = list(dict.fromkeys(values))
                value = metadata.list_to_human(values)
        return value


class Subplot:
    @classmethod
    def from_data(cls, superplot, data, *args, domain=None, crs=None, **kwargs):
        if not isinstance(data, earthkit.data.core.Base):
            data = earthkit.data.from_object(data)

        if domain is None and crs is None:
            try:
                crs = data.projection().to_cartopy_crs()
            except AttributeError:
                pass
        return cls(superplot, *args, domain=domain, crs=crs, **kwargs)

    def __init__(self, superplot, *args, domain=None, crs=None, **kwargs):
        self.domain = domains.parse(domain, crs)

        self.ax = superplot.fig.add_subplot(*args, projection=self.domain.crs, **kwargs)

        if self.domain.bounds is not None:
            self.ax.set_extent(self.domain.bounds, crs=self.domain.crs)

        self.superplot = superplot
        self.layers = []

    @property
    def fig(self):
        return self.superplot.fig

    @property
    def distinct_legend_layers(self):
        """Group layers by style."""
        legend_layers = []
        for layer in self.layers:
            for legend_layer in legend_layers:
                if legend_layer.style == layer.style:
                    break
            else:
                legend_layers.append(layer)
        return legend_layers

    def gridded_scalar(method):
        def wrapper(self, data, x=None, y=None, transform=None, style=None, **kwargs):
            if not isinstance(data, earthkit.data.core.Base):
                data = earthkit.data.from_object(data)
            # - TEMPORARY: in the future all "fields" will be "fieldlists" -
            if isinstance(data, earthkit.data.core.Base) and hasattr(data, "__len__"):
                try:
                    data = data[0]
                except (ValueError, TypeError, AttributeError):
                    pass
            # --------------------------------------------------------------

            is_reduced_gg = False
            if x is not None and y is not None:
                if transform is None:
                    raise ValueError(
                        "you must pass a 'transform' when plotting manually "
                        "with x and y coordinates"
                    )
                values = data
            else:
                if data.metadata("gridType", default=None) == "reduced_gg":
                    is_reduced_gg = True
                    x, y, values = extract_reduced_gg(data, self.domain)
                    kwargs.pop("transform_first", None)
                    transform = self.domain.crs

                else:
                    x, y, values = extract_scalar(data, self.domain)
                if transform is None:
                    try:
                        transform = data.projection().to_cartopy_crs()
                    except AttributeError:
                        transform = ccrs.PlateCarree()

            source_units = data.metadata("units", default=None)

            short_name = metadata.get_metadata(data, "short_name", default="")

            if style is None:
                style_units = None
                if not schema.force_style_units:
                    style_units = kwargs.pop("units", None) or source_units
                style = suggest_style(data, units=style_units)

            values = style.convert_units(values, source_units, short_name=short_name)

            if "transform_first" not in kwargs and not is_reduced_gg:
                kwargs["transform_first"] = self._can_transform_first(method)

            if is_reduced_gg:
                mappable = self._tricontourf(
                    x,
                    y,
                    values,
                    style=style,
                    transform=transform,
                    **kwargs,
                )
            else:
                mappable = method(
                    self,
                    x,
                    y,
                    values,
                    style=style,
                    transform=transform,
                    **kwargs,
                )

            layer = Layer(data, mappable, self, style=style)
            self.layers.append(layer)

            return layer

        return wrapper

    def polygonal(method):
        def wrapper(self, data, column=None, style=None, **kwargs):

            if column is None:
                return self.add_geometries(data, **kwargs)

            values = data.loc[:, column]
            values = [i.item() for i in values.values.flatten()]

            source_units = data.attrs["reduce_attrs"][column].get("units")

            if style is None:
                style_units = None
                if not schema.force_style_units:
                    style_units = kwargs.pop("units", source_units)
                style = suggest_style(data, units=style_units)

            values = style.convert_units(np.array(values), source_units)

            cmap = style.to_kwargs(values)["cmap"]
            norm = style.to_kwargs(values)["norm"]
            for index, (_, row) in enumerate(data.iterrows()):
                color = cmap(norm(values[index]))
                geometry = row["geometry"]
                self.ax.add_geometries(
                    [geometry],
                    crs=ccrs.PlateCarree(),
                    facecolor=color,
                    edgecolor="black",
                )

            x0, y0, x1, y1 = data.total_bounds

            z = [[max(values), min(values)], [max(values), min(values)]]
            x = [x0, x1]
            y = [y0, y1]

            mappable = style.contourf(
                self.ax,
                x,
                y,
                z,
                cmap=cmap,
                norm=norm,
                alpha=0,
                transform=ccrs.PlateCarree(),
            )

            layer = Layer(data, mappable, self, style=style)
            self.layers.append(layer)

            return layer

        return wrapper

    def _can_transform_first(self, method):
        if self.domain._can_transform_first and method.__name__ != "pcolormesh":
            return True
        else:
            return False

    def plot(self, *args, **kwargs):
        data = args[0]

        if isinstance(data, (pd.DataFrame, pd.Series)):
            return self.polygons(*args, **kwargs)
        else:
            return self._plot_gridded_scalar(*args, **kwargs)

    @gridded_scalar
    def _plot_gridded_scalar(self, *args, style=None, **kwargs):
        try:
            return style.plot(self.ax, *args, **kwargs)
        except NotImplementedError:
            return style.contourf(self.ax, *args, **kwargs)

    @gridded_scalar
    def contourf(self, *args, style=None, **kwargs):
        return style.contourf(self.ax, *args, **kwargs)

    def _tricontourf(self, *args, style=None, **kwargs):

        kwargs.pop("transform", None)

        return style.tricontourf(self.ax, *args, **kwargs)

    @polygonal
    def polygons(self, *args, style=None, **kwargs):
        return style.polygons(self.ax, *args, **kwargs)

    @gridded_scalar
    def contour(self, *args, style=None, **kwargs):
        return style.contour(self.ax, *args, **kwargs)

    shaded_contour = contourf

    @gridded_scalar
    def pcolormesh(self, *args, style=None, **kwargs):
        return style.pcolormesh(self.ax, *args, **kwargs)

    @gridded_scalar
    def scatter(self, *args, style=None, **kwargs):
        return style.scatter(self.ax, *args, **kwargs)

    @schema.borders.apply()
    def add_geometries(
        self, shapes, *args, crs=None, labels=False, label_kwargs=None, **kwargs
    ):
        if isinstance(shapes, earthkit.data.core.Base):
            shapes = shapes.to_pandas()

        if crs is None:
            crs = self.domain.crs
            shapes = shapes.to_crs(crs.proj4_init)

        if isinstance(shapes, gpd.geoseries.GeoSeries):
            geometries = shapes
        else:
            geometries = shapes["geometry"]

        if labels:
            label_kwargs = label_kwargs or dict()
            label_kwargs = {
                **dict(
                    ha="center",
                    va="center",
                    bbox=dict(
                        boxstyle="round",
                        ec=(0.2, 0.2, 0.2, 0),
                        fc=(0.3, 0.3, 0.3),
                    ),
                    fontsize=8,
                    weight="bold",
                    color=(0.95, 0.95, 0.95),
                    clip_on=True,
                    clip_box=self.ax.bbox,
                    transform=crs,
                ),
                **label_kwargs,
            }

            if isinstance(labels, str):
                labels = shapes[labels]
            else:
                labels = shapes.iloc[:, 0]

            for i, row in shapes.iterrows():
                try:
                    geometry = max(row.geometry.geoms, key=lambda a: a.area)
                except AttributeError:
                    geometry = row.geometry
                x, y = geometry.representative_point().coords[:][0]
                self.ax.text(x, y, labels[i], **label_kwargs)

        return self.ax.add_geometries(geometries, *args, crs=crs, **kwargs)

    @schema.coastlines.apply()
    def coastlines(self, *args, resolution="auto", **kwargs):
        """Add coastal outlines from the Natural Earth “coastline” collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        return self.ax.coastlines(*args, resolution=resolution, **kwargs)

    @schema.borders.apply()
    def borders(
        self, *args, resolution="auto", labels=False, label_kwargs=None, **kwargs
    ):
        """Add country boundary polygons from Natural Earth.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.BORDERS
        else:
            feature = cfeature.NaturalEarthFeature(
                "cultural", "admin_0_countries", resolution
            )
        if labels:
            label_kwargs = label_kwargs or dict()
            label_kwargs = {
                **dict(
                    ha="center",
                    va="center",
                    bbox=dict(
                        boxstyle="round",
                        ec=(0.2, 0.2, 0.2, 0),
                        fc=(0.3, 0.3, 0.3),
                    ),
                    fontsize=8,
                    weight="bold",
                    color=(0.95, 0.95, 0.95),
                    clip_on=True,
                    clip_box=self.ax.bbox,
                    transform=ccrs.Geodetic(),
                ),
                **label_kwargs,
            }

            label_key = labels if isinstance(labels, str) else "ISO_A2_EH"

            resolution = "110m" if resolution == "auto" else resolution
            shpfilename = shpreader.natural_earth(
                resolution=resolution, category="cultural", name="admin_0_countries"
            )
            reader = shpreader.Reader(shpfilename)
            for record in reader.records():
                name = record.attributes[label_key]
                x = record.attributes["LABEL_X"]
                y = record.attributes["LABEL_Y"]
                self.ax.text(x, y, name, **label_kwargs)
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.land.apply()
    def land(self, *args, resolution="auto", **kwargs):
        """Add land polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.LAND
        else:
            feature = cfeature.NaturalEarthFeature("physical", "land", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.land.apply()
    def ocean(self, *args, resolution="auto", **kwargs):
        """Add ocean polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.OCEAN
        else:
            feature = cfeature.NaturalEarthFeature("physical", "land", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    def stock_img(self, *args, **kwargs):
        self.ax.stock_img(*args, **kwargs)

    @schema.gridlines.apply()
    def gridlines(self, *args, **kwargs):
        """Add gridlines to the map."""
        self._gridlines = self.ax.gridlines(*args, **kwargs)
        return self._gridlines

    @schema.legend.apply()
    def legend(self, *args, **kwargs):
        legends = []
        for layer in self.distinct_legend_layers:
            if layer.style is not None:
                legend = layer.style.legend(
                    self.fig,
                    layer,
                    *args,
                    ax=layer.axes,
                    **kwargs,
                )
            legends.append(legend)
        return legends

    @property
    def _default_title_template(self):
        templates = [layer._default_title_template for layer in self.layers]
        if len(set(templates)) == 1:
            template = templates[0]
        else:
            title_parts = []
            for i, template in enumerate(templates):
                keys = [k for _, k, _, _ in SubplotFormatter().parse(template)]
                for key in set(keys):
                    template = template.replace("{" + key, "{" + key + f"!{i}")
                title_parts.append(template)
            template = metadata.list_to_human(title_parts)
        return template

    @schema.title.apply()
    def title(self, label=None, unique=True, wrap=True, **kwargs):
        if label is None:
            label = self._default_title_template
        label = self.format_string(label, unique)
        plt.sca(self.ax)
        return plt.title(label, wrap=wrap, **kwargs)

    def format_string(self, string, unique=True, grouped=True):
        if not grouped:
            return metadata.list_to_human(
                [LayerFormatter(layer).format(string) for layer in self.layers]
            )
        else:
            return SubplotFormatter(self, unique=unique).format(string)


def extract_scalar(data, domain):
    if hasattr(data, "__len__"):
        try:
            data = data[0]
        except (ValueError, TypeError, AttributeError):
            data = data

    values, points = domain.bbox(data)

    y = points["y"]
    x = points["x"]
    # if data.projection().CF_GRID_MAPPING_NAME == "latitude_longitude":
    #     x[x > 180] -= 360

    return x, y, values


def extract_vector(data, domain):
    x, y, u_values = extract_scalar(data[0], domain)
    _, _, v_values = extract_scalar(data[1], domain)
    return x, y, (u_values, v_values)


def extract_reduced_gg(data, domain):
    values = data.values
    points = data.to_points()

    xy = domain.crs.transform_points(
        ccrs.PlateCarree(),
        points["x"],
        points["y"],
    )
    x = xy[:, 0]
    y = xy[:, 1]

    return x, y, values



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from earthkit.maps.metadata import units


@pytest.mark.skipif(units._NO_CF_UNITS, reason="cf-units is not installed")
def test_are_equal():
    assert units.are_equal("K", "kelvin") is True
    assert units.are_equal("celsius", "kelvin") is False


@pytest.mark.skipif(units._NO_CF_UNITS, reason="cf-units is not installed")
def test_convert():
    assert units.convert(273.15, "kelvin", "celsius") == 0


@pytest.mark.skipif(units._NO_CF_UNITS, reason="cf-units is not installed")
def test_anomaly_equivalence():
    assert units.anomaly_equivalence("celsius") is True
    assert units.anomaly_equivalence("kelvin") is True
    assert units.anomaly_equivalence("fahrenheit") is False


def test_format_units_no_cf_units():
    _no_cf_units = units._NO_CF_UNITS
    units._NO_CF_UNITS = True
    assert units.format_units("celsius") == "$°C$"
    units._NO_CF_UNITS = _no_cf_units


@pytest.mark.skipif(units._NO_CF_UNITS, reason="cf-units is not installed")
def test_format_units_with_cf_units():
    assert units.format_units("celsius") == "$°C$"
    assert units.format_units("s-1") == "${s}^{-1}$"



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np


def auto_range(data, divergence_point=None, n_levels=8):
    """
    Attempt to generate a suitable range of levels for arbitrary input data.

    NOTE: **Experimental!**

    Parameters
    ----------
    data : numpy.ndarray or xarray.DataArray or earthkit.data.core.Base
        The data for which to generate a list of levels.
    divergence_point : float, optional
        If provided, force the levels to be centred on this point. This is
        mostly useful for parameters which use diverging colors in their style,
        such as anomalies.
    n_levels : int, optional
        The number of levels to generate.

    Returns
    -------
    list
    """
    try:
        data = data.to_numpy()
    except AttributeError:
        pass

    min_value = np.nanmin(data)
    max_value = np.nanmax(data)

    if divergence_point is not None:
        max_diff = max(max_value - divergence_point, divergence_point - min_value)
        max_value = divergence_point + max_diff
        min_value = divergence_point - max_diff

    data_range = max_value - min_value

    initial_bin = data_range / n_levels

    magnitude = 10 ** (math.floor(math.log(initial_bin, 10)))
    bin_width = initial_bin - (initial_bin % -magnitude)

    start = min_value - (min_value % magnitude)

    levels = np.arange(
        start,
        start + (bin_width * n_levels) + bin_width,
        bin_width,
    ).tolist()

    while levels[-2] >= max_value:
        levels = levels[:-1]

    return levels


def step_range(data, step, reference=None):
    """
    Generate a range of levels for some data based on a level step and multiple.

    Parameters
    ----------
    data : numpy.ndarray or xarray.DataArray or earthkit.data.core.Base
        The data for which to generate a list of levels.
    step : float
        The step/difference between each level in the desired level range.
    reference : float, optional
        The reference point around which to calibrate the level range. For
        example, if a `step` of 4 is used and a `reference` of 2 is used, then
        the generated levels will be generated as steps of 4 above and below
        the number 2.

    Returns
    -------
    list
    """
    try:
        data = data.to_numpy()
    except AttributeError:
        pass

    if reference is None:
        reference = step

    min_value = np.nanmin(data)
    max_value = np.nanmax(data)

    max_modifier = reference % step
    min_modifier = max_modifier if max_modifier == 0 else step - max_modifier

    min_value = min_value - (min_value % step) - min_modifier

    levels = np.arange(min_value, max_value + step, step).tolist()
    if levels[1] <= np.nanmin(data):
        levels = levels[1:]

    return levels


class Levels:
    """
    Class defining levels to use with a mapping style.

    Parameters
    ----------
    levels : list, optional
        A static list of levels to always use, no matter the input data.
    step : float, optional
        The step/difference between each level in the desired level range.
    reference : float, optional
        The reference point around which to calibrate the level range. For
        example, if a `step` of 4 is used and a `reference` of 2 is used, then
        the generated levels will be generated as steps of 4 above and below
        the number 2.
    divergence_point : float, optional
        If provided, force the levels to be centred on this point. This is
        mostly useful for parameters which use diverging colors in their style,
        such as anomalies.
    """

    @classmethod
    def from_config(cls, config):
        if isinstance(config, str):
            if config.startswith("range"):
                args = (
                    int(i)
                    for i in config.replace("range(", "").replace(")", "").split(",")
                )
                kwargs = {"levels": range(*args)}
        elif isinstance(config, dict):
            kwargs = config
        else:
            kwargs = {"levels": config}
        return cls(**kwargs)

    def __init__(
        self,
        levels=None,
        step=None,
        reference=None,
        divergence_point=None,
    ):
        self._levels = levels
        self._step = step
        self._reference = reference
        self._divergence_point = divergence_point

    def apply(self, data):
        """
        Generate levels specific to some data.

        Parameters
        ----------
        data : numpy.ndarray or xarray.DataArray or earthkit.data.core.Base
            The data for which to generate a list of levels.

        Returns
        -------
        list
        """
        if self._levels is None:
            if self._step is None:
                return auto_range(data, self._divergence_point)
            else:
                return step_range(data, self._step, self._reference)
        return self._levels



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps import schemas


def test_Schema_get():
    schema = schemas.Schema(
        number=1,
        string="hello",
        dictionary={"key": "value"},
    )

    assert schema.number == 1
    assert schema.string == "hello"
    assert schema.dictionary == {"key": "value"}
    assert schema.dictionary.key == "value"


def test_Schema_set():
    schema = schemas.Schema(
        number=1,
    )

    assert schema.number == 1

    schema.number = 2
    assert schema.number == 2

    schema.foo = "bar"
    assert schema.foo == "bar"


def test_Schema_context_manager():
    schema = schemas.Schema(
        number=1,
    )

    assert schema.number == 1

    with schema.set(number=2):
        assert schema.number == 2

    assert schema.number == 1


def test_Schema_recursive_update():
    schema = schemas.Schema(
        dictionary={"foo": "bar", "baz": "qux"},
    )

    assert schema.dictionary.foo == "bar"
    assert schema.dictionary.baz == "qux"

    schema.dictionary.foo = 42
    assert schema.dictionary.foo == 42
    assert schema.dictionary.baz == "qux"


def test_Schema_use():
    schema = schemas.Schema(
        use_preferred_styles=False,
    )

    assert schema.use_preferred_styles is False

    schema.use("ecmwf")

    assert schema.use_preferred_styles is True



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs

from earthkit.maps.domains import crs


def test_from_string():
    assert crs.from_string("Robinson") == ccrs.Robinson()


def test_from_dict():
    kwargs = {
        "name": "LambertAzimuthalEqualArea",
        "central_longitude": 40,
        "central_latitude": -10,
    }
    assert crs.from_dict(kwargs) == (
        ccrs.LambertAzimuthalEqualArea(
            **{k: v for k, v in kwargs.items() if k != "name"}
        )
    )


def test_parse():
    assert crs.parse(None) == ccrs.PlateCarree()

    assert crs.parse("Mollweide") == ccrs.Mollweide()

    kwargs = {
        "name": "LambertAzimuthalEqualArea",
        "central_longitude": 40,
        "central_latitude": -10,
    }
    assert crs.parse(kwargs) == (
        ccrs.LambertAzimuthalEqualArea(
            **{k: v for k, v in kwargs.items() if k != "name"}
        )
    )



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt

from earthkit.maps import metadata, styles
from earthkit.maps.schemas import schema
from earthkit.maps.styles import auto, colors, legends, levels

__all__ = [
    "colors",
    "legends",
    "levels",
    "auto",
    "Style",
]


class Style:
    """
    Class describing a mapping style.

    Parameters
    ----------
    colors : str or list or matplotlib.colors.Colormap, optional
        The colors to be used in this `Style`. This can be a
        `named matplotlib colormap
        <https://matplotlib.org/stable/gallery/color/colormap_reference.html>`__,
        a list of colors (as named CSS4 colors, hexadecimal colors or three
        (four)-element lists of RGB(A) values), or a pre-defined matplotlib
        colormap object. If not provided, the default colormap of the active
        `schema` will be used.
    levels : list or earthkit.maps.styles.levels.Levels, optional
        The levels to use in this `Style`. This can be a list of specific
        levels, or an earthkit `Levels` object. If not provided, some suitable
        levels will be generated automatically (experimental!).
    gradients : list, optional
        The number of colors to insert between each level in `levels`. If None,
        one color level will be inserted between each level.
    normalize : bool, optional
        If `True` (default), then the colors will be normalized over the level
        range.
    units : str, optional
        The units in which the levels are defined. If this `Style` is used with
        data not using the given units, then a conversion will be attempted;
        any data incompatible with these units will not be able to use this
        `Style`. If `units` are not provided, then data plotted using this
        `Style` will remain in their original units.
        Note that passing `units` requires `cf_units` to be installed.
    units_label : str, optional
        The label to use in titles and legends to represent the units of the
        data.
    legend_style : str, optional
        The style of legend to use by default with this style. Must be one of
        `colorbar` (default), `disjoint`, `histogram`, or `None` (no legend).
    bin_labels : list, optional
        A list of categorical labels for each bin in the legend.
    """

    @classmethod
    def from_dict(cls, kwargs):
        style_type = kwargs.pop("type")
        if "levels" in kwargs:
            kwargs["levels"] = levels.Levels.from_config(kwargs["levels"])
        return getattr(styles, style_type)(**kwargs)

    # TODO
    # @classmethod
    # def from_magics_style(cls, magics_style):
    #     pass

    def __init__(
        self,
        colors=schema.cmap,
        levels=None,
        gradients=None,
        normalize=True,
        units=None,
        units_label=None,
        legend_style="colorbar",
        legend_kwargs=None,
        bin_labels=None,
        ticks=None,
        **kwargs,
    ):
        self._colors = colors
        self._levels = (
            levels
            if isinstance(levels, styles.levels.Levels)
            else styles.levels.Levels(levels)
        )
        self.normalize = normalize
        self.gradients = gradients

        if units is not None and metadata.units._NO_CF_UNITS:
            warnings.warn(
                "You must have cf-units installed to use unit conversion "
                "features; since no cf-units installation was found, no units "
                "will be applied to this style"
            )
            units = None
        self._units = units
        self._units_label = units_label

        self._legend_style = legend_style
        if self._legend_style == "None":
            self._legend_style = None

        self._bin_labels = bin_labels
        self._legend_kwargs = legend_kwargs or dict()
        if ticks is not None:
            self._legend_kwargs["ticks"] = ticks

        self._kwargs = kwargs

    # TODO
    # def to_yaml(self):
    #     pass

    # TODO
    # def to_magics_style(self):
    #     pass

    def levels(self, data=None):
        """
        Generate levels specific to some data.

        Parameters
        ----------
        data : numpy.ndarray or xarray.DataArray or earthkit.data.core.Base
            The data for which to generate a list of levels.

        Returns
        -------
        list
        """
        if data is None:
            if self._levels._levels is not None:
                return self._levels._levels
            else:
                raise ValueError(
                    "this style uses dynamic levels; include the `data` "
                    "argument to generate levels"
                )
        return self._levels.apply(data)

    @property
    def extend(self):
        """Convenience access to 'extend' kwarg."""
        return self._kwargs.get("extend")

    @property
    def units(self):
        """Formatted units for use in figure text."""
        if self._units_label is not None:
            return self._units_label
        elif self._units is not None:
            return metadata.units.format_units(self._units)

    def convert_units(self, values, source_units, short_name=""):
        """
        Convert some values from their source units to this `Style`'s units.

        Parameters
        ----------
        values : numpy.ndarray
            The values to convert from their source units to this `Style`'s
            units.
        source_units : str or cf_units.Unit
            The source units of the given values.
        short_name : str, optional
            The short name of the variable, which is used to make extra
            assumptions about the data's unit covnersion (for example,
            temperature anomalies need special consideration when converting
            between Celsius and Kelvin).
        """
        if self._units is None or source_units is None:
            return values

        # For temperature anomalies we do not want to convert values, just
        # change the units string
        if "anomaly" in short_name.lower() and metadata.units.anomaly_equivalence(
            source_units
        ):
            return values

        return metadata.units.convert(values, source_units, self._units)

    def to_matplotlib_kwargs(self, data):
        """
        Generate matplotlib arguments required for plotting data in this `Style`.

        Parameters
        ----------
        data : numpy.ndarray
            The data to be plotted using this `Style`.
        """
        levels = self.levels(data)

        if self.gradients is not None:
            self._legend_kwargs.setdefault(
                "ticks", None
            )  # Let matplotlib auto-generate ticks
            return colors.gradients(
                levels,
                self._colors,
                self.gradients,
                self.normalize,
                **self._kwargs,
            )

        cmap, norm = styles.colors.cmap_and_norm(
            self._colors,
            levels,
            self.normalize,
            self.extend,
        )

        return {
            **{"cmap": cmap, "norm": norm, "levels": levels},
            **self._kwargs,
        }

    def to_contourf_kwargs(self, data):
        """
        Generate `contourf` arguments required for plotting data in this `Style`.

        Parameters
        ----------
        data : numpy.ndarray
            The data to be plotted using this `Style`.
        """
        kwargs = self.to_matplotlib_kwargs(data)
        kwargs.pop("linewidths", None)
        return kwargs

    def to_contour_kwargs(self, data):
        """
        Generate `contour` arguments required for plotting data in this `Style`.

        Parameters
        ----------
        data : numpy.ndarray
            The data to be plotted using this `Style`.
        """
        return self.to_matplotlib_kwargs(data)

    def to_pcolormesh_kwargs(self, data):
        """
        Generate `pcolormesh` arguments required for plotting data in this `Style`.

        Parameters
        ----------
        data : numpy.ndarray
            The data to be plotted using this `Style`.
        """
        kwargs = self.to_matplotlib_kwargs(data)
        kwargs.pop("levels", None)
        kwargs.pop("transform_first", None)
        kwargs.pop("extend", None)
        return kwargs

    def to_scatter_kwargs(self, data):
        """
        Generate `scatter` arguments required for plotting data in this `Style`.

        Parameters
        ----------
        data : numpy.ndarray
            The data to be plotted using this `Style`.
        """
        kwargs = self.to_matplotlib_kwargs(data)
        kwargs.pop("levels", None)
        return kwargs

    def plot(self, *args, **kwargs):
        """Plot the data using the `Style`'s defaults."""
        return self.contourf(*args, **kwargs)

    def contourf(self, ax, x, y, values, *args, **kwargs):
        """
        Plot shaded contours using this `Style`.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to plot the data.
        x : numpy.ndarray
            The x coordinates of the data to be plotted.
        y : numpy.ndarray
            The y coordinates of the data to be plotted.
        values : numpy.ndarray
            The values of the data to be plotted.
        **kwargs
            Any additional arguments accepted by `matplotlib.axes.Axes.contourf`.
        """
        kwargs = {**self.to_contourf_kwargs(values), **kwargs}
        return ax.contourf(x, y, values, *args, **kwargs)

    def barbs(self, ax, x, y, u, v, *args, **kwargs):
        return ax.barbs(x, y, u, v, *args, **kwargs)

    def tricontourf(self, ax, x, y, values, *args, **kwargs):
        """
        Plot triangulated shaded contours using this `Style`.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to plot the data.
        x : numpy.ndarray
            The x coordinates of the data to be plotted.
        y : numpy.ndarray
            The y coordinates of the data to be plotted.
        values : numpy.ndarray
            The values of the data to be plotted.
        **kwargs
            Any additional arguments accepted by `matplotlib.axes.Axes.tricontourf`.
        """
        kwargs = {**self.to_contourf_kwargs(values), **kwargs}
        return ax.tricontourf(x, y, values, *args, **kwargs)

    def contour(self, ax, x, y, values, *args, **kwargs):
        """
        Plot line contours using this `Style`.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to plot the data.
        x : numpy.ndarray
            The x coordinates of the data to be plotted.
        y : numpy.ndarray
            The y coordinates of the data to be plotted.
        values : numpy.ndarray
            The values of the data to be plotted.
        **kwargs
            Any additional arguments accepted by `matplotlib.axes.Axes.contour`.
        """
        kwargs = {**self.to_contour_kwargs(values), **kwargs}
        return ax.contour(x, y, values, *args, **kwargs)

    def pcolormesh(self, ax, x, y, values, *args, **kwargs):
        """
        Plot a pcolormesh using this `Style`.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to plot the data.
        x : numpy.ndarray
            The x coordinates of the data to be plotted.
        y : numpy.ndarray
            The y coordinates of the data to be plotted.
        values : numpy.ndarray
            The values of the data to be plotted.
        **kwargs
            Any additional arguments accepted by `matplotlib.axes.Axes.pcolormesh`.
        """
        kwargs.pop("transform_first", None)
        kwargs = {**self.to_pcolormesh_kwargs(values), **kwargs}
        return ax.pcolormesh(x, y, values, *args, **kwargs)

    def scatter(self, ax, x, y, values, s=3, *args, **kwargs):
        """
        Plot a scatter plot using this `Style`.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to plot the data.
        x : numpy.ndarray
            The x coordinates of the data to be plotted.
        y : numpy.ndarray
            The y coordinates of the data to be plotted.
        values : numpy.ndarray
            The values of the data to be plotted.
        **kwargs
            Any additional arguments accepted by `matplotlib.axes.Axes.scatter`.
        """
        kwargs.pop("transform_first", None)
        kwargs = {**self.to_scatter_kwargs(values), **kwargs}
        return ax.scatter(x, y, c=values, s=s, *args, **kwargs)

    def values_to_colors(self, values, data=None):
        """
        Convert a value or list of values to colors based on this `Style`.

        Parameters
        ----------
        values : float or list of floats
            The values to convert to colors on this `Style`'s color scale.
        """
        mpl_kwargs = self.to_matplotlib_kwargs(data=data)
        cmap = mpl_kwargs["cmap"]
        norm = mpl_kwargs["norm"]
        return cmap(norm(values))

    def legend(self, *args, **kwargs):
        """Create the default legend for this `Style`."""
        if self._legend_style is None:
            return

        try:
            method = getattr(self, self._legend_style)
        except AttributeError:
            raise AttributeError(f"invalid legend type '{self._legend_style}'")

        return method(*args, **kwargs)

    def colorbar(self, *args, **kwargs):
        return styles.legends.colorbar(*args, **kwargs)

    def disjoint(self, *args, **kwargs):
        return styles.legends.disjoint(*args, **kwargs)

    def save_legend_graphic(
        self, filename="legend.png", data=None, transparent=True, **kwargs
    ):
        """
        Save a standalone image of the legend associated with this `Style`.

        Parameters
        ----------
        filename : str
            The name of the image to save.
        data : earthkit.data.core.Base, optional
            It can sometimes be useful to pass some data in order to
            automatically generate legend labels or color ranges, depending on
            the `Style`.
        """
        x = None
        y = None

        if data is None:
            data = [[1, 2], [3, 4]]
            x = [[1, 2], [3, 4]]
            y = [[1, 2], [3, 4]]
            kwargs["label"] = kwargs.get("label", "")

        backend = mpl.get_backend()
        mpl.use("Agg")

        try:
            getattr(self, f"_save_{self._legend_style}_graphic")(
                data, x, y, filename, transparent, kwargs
            )
        finally:
            mpl.use(backend)

    def _save_colorbar_graphic(self, data, x, y, filename, transparent, kwargs):
        from earthkit.maps import Chart

        chart = Chart()
        chart.contourf(data, x=x, y=y, style=self)

        legend = chart.legend(**kwargs)[0]

        chart.fig.canvas.draw()
        bbox = legend.ax.get_window_extent().transformed(
            chart.fig.dpi_scale_trans.inverted()
        )

        title_bbox = legend.ax.xaxis.label.get_window_extent().transformed(
            chart.fig.dpi_scale_trans.inverted()
        )

        x, y = chart.fig.get_size_inches()

        xmod, ymod = (
            (0.05, 0.01) if legend.orientation == "horizontal" else (0.01, 0.05)
        )

        bbox.x0 = min(bbox.x0, title_bbox.x0) - x * xmod
        bbox.x1 = max(bbox.x1, title_bbox.x1) + x * xmod
        bbox.y0 = min(bbox.y0, title_bbox.y0) - y * ymod
        bbox.y1 = max(bbox.y1, title_bbox.y1) + y * ymod

        plt.savefig(filename, dpi="figure", bbox_inches=bbox, transparent=transparent)

    def _save_disjoint_graphic(self, data, x, y, filename, transparent, kwargs):
        from earthkit.maps import Chart

        chart = Chart()
        chart.contourf(data, x=x, y=y, style=self)

        legend = chart.legend(**kwargs)[0]

        chart.fig.canvas.draw()
        fig = legend.figure
        fig.canvas.draw()
        bbox = legend.get_window_extent().transformed(fig.dpi_scale_trans.inverted())

        plt.savefig(filename, dpi="figure", bbox_inches=bbox, transparent=transparent)


class Contour(Style):
    def __init__(
        self,
        colors=None,
        line_colors=None,
        labels=False,
        label_kwargs=None,
        interpolate=True,
        **kwargs,
    ):
        super().__init__(colors=colors, **kwargs)
        self._line_colors = line_colors
        self.labels = labels
        self._label_kwargs = label_kwargs or dict()
        self._interpolate = interpolate
        self._kwargs["linewidths"] = kwargs.get("linewidths", 0.5)

    def plot(self, *args, **kwargs):
        if self._colors is not None:
            if self._interpolate:
                return self.contourf(*args, **kwargs)
            else:
                return self.pcolormesh(*args, **kwargs)
        else:
            return self.contour(*args, **kwargs)

    def to_contour_kwargs(self, data):
        levels = self.levels(data)
        cmap, norm = styles.colors.cmap_and_norm(
            self._line_colors,
            levels,
            self.normalize,
            self.extend,
        )

        return {
            **{"cmap": cmap, "norm": norm, "levels": levels},
            **self._kwargs,
        }

    def contourf(self, ax, x, y, values, *args, **kwargs):
        mappable = super().contourf(ax, x, y, values, *args, **kwargs)
        if self._line_colors is not None:
            self.contour(ax, x, y, values, *args, **kwargs)
        return mappable

    def contour(self, *args, **kwargs):
        mappable = super().contour(*args, **kwargs)

        if self.labels:
            self.contour_labels(mappable, **self._label_kwargs)

        return mappable

    def contour_labels(
        self,
        mappable,
        label_fontsize=7,
        label_colors=None,
        label_frequency=1,
        label_background=None,
        label_fmt=None,
    ):
        clabels = mappable.axes.clabel(
            mappable,
            mappable.levels[0::label_frequency],
            inline=True,
            fontsize=label_fontsize,
            colors=label_colors,
            fmt=label_fmt,
            inline_spacing=2,
        )
        if label_background is not None:
            for label in clabels:
                label.set_backgroundcolor(label_background)

        return clabels


class Hatched(Contour):
    def __init__(self, *args, hatches=".", background_colors=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.hatches = hatches
        self._foreground_colors = self._colors
        self._colors = background_colors or [(0, 0, 0, 0)]

    def contourf(self, *args, **kwargs):
        mappable = super().contourf(*args, hatches=self.hatches, **kwargs)

        line_colors = colors.expand(self._foreground_colors, mappable.levels)

        for i, collection in enumerate(mappable.collections):
            collection.set_edgecolor(line_colors[i])
            collection.set_linewidth(0)

        return mappable

    def colorbar(self, *args, **kwargs):
        colorbar = super().colorbar(*args, **kwargs)

        levels = colorbar.mappable.levels

        line_colors = colors.expand(self._foreground_colors, levels)
        for i, artist in enumerate(colorbar.solids_patches):
            artist.set_edgecolor(line_colors[i])

        return colorbar

    def disjoint(self, layer, *args, **kwargs):
        legend = super().disjoint(layer, *args, **kwargs)

        line_colors = colors.expand(self._foreground_colors, layer.mappable.levels)

        for color, artist in zip(line_colors, legend.get_patches()):
            artist.set_edgecolor(color)
            artist.set_linewidth(0.0)

        return legend


DEFAULT_STYLE = Style()



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps import utils


def test_recursive_dict_update():
    original_dict = {"test": {"a": 1, "b": 2}, "foo": "bar", "qux": 42}
    update_dict = {"test": {"b": 3, "c": 4}, "foo": "baz"}

    assert utils.recursive_dict_update(original_dict, update_dict) == {
        "test": {"a": 1, "b": 3, "c": 4},
        "foo": "baz",
        "qux": 42,
    }


def test_list_to_human():
    assert utils.list_to_human(["a"]) == "a"
    assert utils.list_to_human(["a", "b"]) == "a and b"
    assert utils.list_to_human(["a", "b", "c"]) == "a, b and c"
    assert utils.list_to_human(["a", "b", "c"], oxford_comma=True) == "a, b, and c"
    assert utils.list_to_human(["a", "b", "c"], conjunction="et") == "a, b et c"



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import earthkit.data
import matplotlib.pyplot as plt
import numpy as np

from earthkit.maps import domains, utils
from earthkit.maps.charts import layouts
from earthkit.maps.charts.layers import LayerGroup
from earthkit.maps.charts.subplots import Subplot
from earthkit.maps.metadata.formatters import ChartFormatter
from earthkit.maps.schemas import schema


class Chart:

    """
    The top-level container for all plot elements.

    Parameters
    ----------
    domain : str or list or earthkit.maps.domains.Domain, optional
        A named domain, or a list of domain extents in the order
        `[x_min, x_max, y_min, y_max]`, describing the domain tp be used by all
        subplots in the `Chart`.
        If not provided, an automatic domain will be created based on the
        extents of any data plotted on the `Chart` (global if no data is
        plotted).
    domain_crs : str or cartopy.crs.CRS, optional
        The coordinate reference system in which the `domain` extents are
        defined. By default, a regular equirectangular latitude-longitude CRS
        (Plate Carrée) is assumed.
    crs : str or cartopy.crs.CRS, optional
        The coordinate reference system to use for all subplots on this `Chart`.
        If not provided, a suitable coordinate-reference system will be created
        based on the `Chart`'s `domain`.
    rows : int, optional
        The number of rows to use for subplots on this `Chart`.
    columns : int, optional
        The number of columns to use for subplots on this `Chart`.
    """

    MAX_COLS = 8
    """
    The default maximum number of subplot columns - can be overridded by
    instantiating a `Chart` with the `columns` argument.
    """

    @classmethod
    def from_gridspec(cls, gridspec, *args, **kwargs):
        """
        Instantiate an earthkit Chart from a matplotlib gridspec.

        Parameters
        ----------
        gridspec : matplotlib.gridspec.GridSpec
            A matplotlib grid layout which determines subplot layout.
        *args, **kwargs

        """
        obj = cls(*args, **kwargs)
        obj._gridspec = gridspec
        return obj

    def __init__(self, domain=None, domain_crs=None, crs=None, rows=None, columns=None):
        self._custom_domain = False
        if isinstance(domain, domains.Domain):
            self._domain = str(domain)
            self._domain_crs = domain.crs
            self._crs = domain.crs
            self.domain = domain
            self._custom_domain = True
        else:
            self._domain = domain
            self._domain_crs = domain_crs
            self._crs = crs
            self.domain = domains.parse(domain, crs)

        self._fig = None
        self._gridspec = None

        self._rows = rows
        self._columns = columns

        self.subplots = []
        self.__subplots_generator = None

        self._queue = []

    def __len__(self):
        return len(self.subplots)

    def __getitem__(self, i):
        if not isinstance(i, slice):
            while i + 1 > len(self.subplots):
                self.add_subplot()
        return self.subplots[i]

    @property
    def fig(self):
        """The `Chart`'s underlying matplotlib `Figure` object."""
        if self._fig is None:
            self._fig = plt.figure(
                figsize=schema.figsize, constrained_layout=True, dpi=schema.dpi
            )
        return self._fig

    @property
    def gridspec(self):
        """The `Chart`'s underlying matplotlib `GridSpec` object."""
        if self._gridspec is None:
            self._gridspec = self.fig.add_gridspec(*self.shape)
        return self._gridspec

    @property
    def rows(self):
        """The number of rows in the `Chart`'s subplot layout."""
        if self._rows is None:
            if len(self) == 0:
                raise ValueError("cannot get rows from empty figure")
            self._rows = layouts.rows_cols(
                len(self), cols=self._columns, max_cols=self.MAX_COLS
            )[0]
        return self._rows

    @property
    def columns(self):
        """The number of columns in the `Chart`'s subplot layout."""
        if self._columns is None:
            if len(self) == 0:
                raise ValueError("cannot get cols from empty figure")
            self._columns = layouts.rows_cols(
                len(self), rows=self._rows, max_cols=self.MAX_COLS
            )[1]
        return self._columns

    @property
    def shape(self):
        """The shape of the `Chart`'s subplot layout."""
        return self.rows, self.columns

    @property
    def _subplots_generator(self):
        if self.__subplots_generator is None:
            self.__subplots_generator = (
                (row, col)
                for row, col in itertools.product(
                    range(self.gridspec.nrows),
                    range(self.gridspec.ncols),
                )
            )
        return self.__subplots_generator

    def add_subplot(
        self, *args, data=None, domain=None, crs=None, row=None, column=None, **kwargs
    ):
        """
        Add a subplot within the `Chart`'s subplot layout.

        Parameters
        ----------
        data : earthkit.data.core.Base, optional
            If provided, generate a subplot with a domain and CRS based on the
            extents of the input data.
        domain : str or list or earthkit.maps.domains.Domain, optional
            A named domain, or a list of domain extents in the order
            `[x_min, x_max, y_min, y_max]`, describing the domain tp be used by
            all subplots in the `Chart`.
            If not provided, the top-level `Chart`'s `domain` will be used.
        domain_crs : str or cartopy.crs.CRS, optional
            The coordinate reference system in which the `domain` extents are
            defined. By default, a regular equirectangular latitude-longitude
            CRS (Plate Carrée) is assumed.
        crs : str or cartopy.crs.CRS, optional
            The coordinate reference system to use for all subplots on this
            `Chart`. If not provided, a suitable coordinate-reference system
            will be created based on the `Chart`'s `domain`.
        rows : int, optional
            The row position at which to insert this subplot.
        columns : int, optional
            The column position at which to insert this subplot.
        """
        if row is None and column is None:
            row, column = next(self._subplots_generator)

        if domain is None and crs is None:
            domain = self.domain if self._custom_domain else self._domain
            domain_crs = self._domain_crs
            crs = self._crs

        if not args:
            args = (self.gridspec[row, column],)

        if data is not None:
            subplot = Subplot.from_data(
                self,
                data,
                *args,
                domain=domain,
                domain_crs=domain_crs,
                crs=crs,
                **kwargs,
            )
        else:
            subplot = Subplot(
                self, *args, domain=domain, domain_crs=domain_crs, crs=crs, **kwargs
            )

        self.subplots.append(subplot)
        return subplot

    def distinct_legend_layers(self, subplots=None):
        """
        Get a list of layers with distinct styles.

        Parameters
        ----------
        subplots : list, optional
            If provided, only these subplots will be considered when identifying
            unique styles.
        """
        if subplots is None:
            subplots = self.subplots

        subplot_layers = [subplot.distinct_legend_layers for subplot in subplots]
        subplot_layers = [item for sublist in subplot_layers for item in sublist]

        groups = []
        for layer in subplot_layers:
            for i in range(len(groups)):
                if groups[i][0].style == layer.style:
                    groups[i].append(layer)
                    break
            else:
                groups.append([layer])

        groups = [LayerGroup(layers) for layers in list(groups)]

        return groups

    def _defer(method):
        """Defer a method's execution until this superplot has subplots."""

        def wrapper(self, *args, **kwargs):
            if not self.subplots:
                self._queue.append((method, args, kwargs))
            else:
                return method(self, *args, **kwargs)

        return wrapper

    def _expand_rows_cols(method):
        def wrapper(self, data, *args, **kwargs):
            if not isinstance(data, (earthkit.data.core.Base, list, np.ndarray)):
                data = earthkit.data.from_object(data)
            if not isinstance(data, earthkit.data.core.Base) or not hasattr(
                data, "__len__"
            ):
                data = [data]

            if not self.subplots:
                num_subplots = len(data)
                self._rows, self._columns = layouts.rows_cols(
                    num_subplots,
                    rows=self._rows,
                    cols=self._columns,
                    max_cols=self.MAX_COLS,
                )

            layers = []
            for i, field in enumerate(data):
                if i + 1 > len(self.subplots):
                    subplot = self.add_subplot(data=field)
                else:
                    subplot = self.subplots[i]
                layer = getattr(subplot, method.__name__)(field, *args, **kwargs)
                layers.append(layer)

            return LayerGroup(layers)

        return wrapper

    def plot(self, data, *args, **kwargs):
        """
        Plot some data on this `Chart`.

        Parameters
        ----------
        data : earthkit.data.core.Base or numpy.ndarray or xarray.Dataset
            The data to plot on the chart.
        x : numpy.ndarray, optional
            The values of the geospatial x coordinate of the data. Only required
            if input data is a raw numpy array of values.
        y : numpy.ndarray, optional
            The values of the geospatial y coordinate of the data. Only required
            if input data is a raw numpy array of values.
        transform : cartopy.crs.CCRS, optional
            The CRS of the source x and y values.
        style : earthkit.maps.styles.Style, optional
            The style to use for this visualisation. If no style is passed,
            earthkit-maps will attempt to choose a suitable style from its
            built-in styles library.

        **kwargs : dict, optional
            Extra arguments to pass to the underlying matplotlib plotting
            method.
        """
        if data.__class__.__name__ in ("DataFrame", "Series"):
            raise NotImplementedError(
                "Plotting pandas DataFrames and Series is an upcoming feature"
            )
        else:
            return self._plot_gridded_scalar(data, *args, **kwargs)

    @_expand_rows_cols
    def _plot_gridded_scalar(self, *args, **kwargs):
        """
        Plot some data.

        Parameters
        ----------
        data : earthkit.data.core.Base or numpy.ndarray or xarray.Dataset
            The data to plot on the chart.
        x : numpy.ndarray, optional
            The x values of the geos
        y : numpy.ndarray, optional
            The y values of the geos
        transform : cartopy.crs.CCRS, optional
        style : earthkit.maps.styles.Style, optional

        **kwargs : dict, optional
            Extra arguments to pass to the underlying matplotlib plotting
            method.
        """
        pass

    @_expand_rows_cols
    def contourf(self, *args, **kwargs):
        """
        Plot some data on this `Chart` using the matplotlib `contourf` method.

        Parameters
        ----------
        data : earthkit.data.core.Base or numpy.ndarray or xarray.Dataset
            The data to plot on the chart.
        x : numpy.ndarray, optional
            The values of the geospatial x coordinate of the data. Only required
            if input data is a raw numpy array of values.
        y : numpy.ndarray, optional
            The values of the geospatial y coordinate of the data. Only required
            if input data is a raw numpy array of values.
        transform : cartopy.crs.CCRS, optional
            The CRS of the source x and y values.
        style : earthkit.maps.styles.Style, optional
            The style to use for this visualisation. If no style is passed,
            earthkit-maps will attempt to choose a suitable style from its
            built-in styles library.

        **kwargs : dict, optional
            Extra arguments to pass to the underlying matplotlib plotting
            method.
        """
        pass

    @_expand_rows_cols
    def contour(self, *args, **kwargs):
        """
        Plot some data on this `Chart` using the matplotlib `contour` method.

        Parameters
        ----------
        data : earthkit.data.core.Base or numpy.ndarray or xarray.Dataset
            The data to plot on the chart.
        x : numpy.ndarray, optional
            The values of the geospatial x coordinate of the data. Only required
            if input data is a raw numpy array of values.
        y : numpy.ndarray, optional
            The values of the geospatial y coordinate of the data. Only required
            if input data is a raw numpy array of values.
        transform : cartopy.crs.CCRS, optional
            The CRS of the source x and y values.
        style : earthkit.maps.styles.Style, optional
            The style to use for this visualisation. If no style is passed,
            earthkit-maps will attempt to choose a suitable style from its
            built-in styles library.

        **kwargs : dict, optional
            Extra arguments to pass to the underlying matplotlib plotting
            method.
        """
        pass

    @_expand_rows_cols
    def pcolormesh(self, *args, **kwargs):
        """
        Plot some data on this `Chart` using the matplotlib `pcolormesh` method.

        Parameters
        ----------
        data : earthkit.data.core.Base or numpy.ndarray or xarray.Dataset
            The data to plot on the chart.
        x : numpy.ndarray, optional
            The values of the geospatial x coordinate of the data. Only required
            if input data is a raw numpy array of values.
        y : numpy.ndarray, optional
            The values of the geospatial y coordinate of the data. Only required
            if input data is a raw numpy array of values.
        transform : cartopy.crs.CCRS, optional
            The CRS of the source x and y values.
        style : earthkit.maps.styles.Style, optional
            The style to use for this visualisation. If no style is passed,
            earthkit-maps will attempt to choose a suitable style from its
            built-in styles library.

        **kwargs : dict, optional
            Extra arguments to pass to the underlying matplotlib plotting
            method.
        """
        pass

    @_expand_rows_cols
    def scatter(self, *args, **kwargs):
        """
        Plot some data on this `Chart` using the matplotlib `scatter` method.

        Parameters
        ----------
        data : earthkit.data.core.Base or numpy.ndarray or xarray.Dataset
            The data to plot on the chart.
        x : numpy.ndarray, optional
            The values of the geospatial x coordinate of the data. Only required
            if input data is a raw numpy array of values.
        y : numpy.ndarray, optional
            The values of the geospatial y coordinate of the data. Only required
            if input data is a raw numpy array of values.
        transform : cartopy.crs.CCRS, optional
            The CRS of the source x and y values.
        style : earthkit.maps.styles.Style, optional
            The style to use for this visualisation. If no style is passed,
            earthkit-maps will attempt to choose a suitable style from its
            built-in styles library.

        **kwargs : dict, optional
            Extra arguments to pass to the underlying matplotlib plotting
            method.
        """
        pass

    @_expand_rows_cols
    def barbs(self, *args, **kwargs):
        pass

    @_defer
    def coastlines(self, *args, **kwargs):
        """
        Add coastal outlines from the Natural Earth “coastline” collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.coastlines(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def borders(self, *args, **kwargs):
        """
        Add country boundaries from the Natural Earth collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.borders(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def states_provinces(self, *args, **kwargs):
        """
        Add state/province boundaries from the Natural Earth collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.states_provinces(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def land(self, *args, **kwargs):
        """
        Add shaded land polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.land(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def ocean(self, *args, **kwargs):
        """
        Add shaded ocean polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.ocean(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def rivers(self, *args, **kwargs):
        """
        Add river polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.rivers(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def lakes(self, *args, **kwargs):
        """
        Add shaded lake polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.lakes(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def urban_areas(self, *args, **kwargs):
        """
        Add urban areas polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: str, optional
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.urban_areas(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def shapes(self, *args, **kwargs):
        """
        Add shapes to each subplot in the `Chart`.

        Parameters
        ----------
        shapes : str or cartopy.io.shapereader.Reader
            The shapes to add to each subplot. Either a path to a shapefile, or
            a cartopy shape reader object.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.shapes(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def cities(self, *args, **kwargs):
        """
        Add cities to each subplot in the `Chart`.

        Parameters
        ----------
        density: str, optional
            The density of cities to plot. One of "low", "medium" or "high".
        capitals_only : bool, optional
            If `True`, only capital cities will be plotted (default `False`).
        adjust_labels: bool, optional, EXPERIMENTAL
            If `True`, an algorithm will be applied which attempts to adjust
            city labels to minimise overlapping. **This can be very slow**.

        Note
        ----
            Matplotlib keyword arguments can be used when drawing the feature.
            This allows standard Matplotlib control over aspects such as
            'facecolor', 'alpha', etc.
        """
        return [subplot.cities(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def image(self, *args, **kwargs):
        return [subplot.image(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def stock_img(self, *args, **kwargs):
        return [subplot.stock_img(*args, **kwargs) for subplot in self.subplots]

    @_defer
    def gridlines(self, *args, **kwargs):
        """
        Add gridlines to the chart.

        Parameters
        ----------
        xstep : float, optional
            The step/difference between each x gridline.
        xref : float, optional
            The reference point around which to calibrate the x level range.
        ystep : float, optional
            The step/difference between each y gridline.
        yref : float, optional
            The reference point around which to calibrate the y level range.

        Note
        ----
            Any keyword arguments accepted by cartopy's `gridlines` method can
            be used when adding gridlines.
        """
        return [subplot.gridlines(*args, **kwargs) for subplot in self.subplots]

    def format_string(self, string, unique=True, grouped=True):
        if not grouped:
            results = [
                subplot.format_string(string, unique, grouped)
                for subplot in self.subplots
            ]
            result = utils.list_to_human(results)
        else:
            result = ChartFormatter(self.subplots, unique=unique).format(string)
        return result

    def _release_queue(self):
        while len(self._queue):
            method, args, kwargs = self._queue.pop(0)
            method(self, *args, **kwargs)

    @schema.legend.apply()
    def legend(self, *args, subplots=None, location=None, **kwargs):
        """
        Add legends to the chart for all subplots.

        Parameters
        ----------

        """
        legends = []

        anchor = None
        non_cbar_layers = []
        for i, layer in enumerate(self.distinct_legend_layers(subplots)):
            if isinstance(location, (list, tuple)):
                loc = location[i]
            else:
                loc = location
            if layer.style is not None:
                legend = layer.style.legend(
                    layer,
                    *args,
                    location=loc,
                    **kwargs,
                )
            if legend.__class__.__name__ != "Colorbar":
                non_cbar_layers.append(layer)
            else:
                anchor = layer.axes[0].get_anchor()
            legends.append(legend)

        if anchor is not None:
            for layer in non_cbar_layers:
                for ax in layer.axes:
                    ax.set_anchor(anchor)

        return legends

    @property
    def _default_title_template(self):
        return self.subplots[0]._default_title_template

    @_defer
    @schema.title.apply()
    def title(self, label=None, unique=True, grouped=True, y=None, **kwargs):
        """
        Add a top-level title to the chart.

        Parameters
        ----------
        label : str, optional
            The text to use in the title. This text can include format keys
            surrounded by `{}` curly brackets, which will extract metadata from
            your plotted data layers.
        unique : bool, optional
            If True, format keys which are uniform across subplots/layers will
            produce a single result. For example, if all data layers have the
            same `variable_name`, only one variable name will appear in the
            title.
            If False, each format key will evaluate to a list of values found
            across subplots/layers.
        grouped : bool, optional
            If True, a single title will be generated to represent all data
            layers, with each format key evaluating to a list where layers
            differ - e.g. `"{variable} at {time}"` might be evaluated to
            `"temperature and wind at 2023-01-01 00:00".
            If False, the title will be duplicated by the number of subplots/
            layers - e.g. `"{variable} at {time}"` might be evaluated to
            `"temperature at 2023-01-01 00:00 and wind at 2023-01-01 00:00".
        kwargs : dict, optional
            Keyword argument to matplotlib.pyplot.suptitle (see
            https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.suptitle.html#matplotlib-pyplot-suptitle
            ).
        """
        if label is None:
            label = self._default_title_template
        label = self.format_string(label, unique, grouped)

        # if y is None:
        #     y = 0
        #     for subplot in self:
        #         y = max(y, subplot.ax.get_position().y1)
        #         if subplot.ax.title.get_text():
        #             y += 0.05
        #     kwargs["va"] = kwargs.get("va", kwargs.get("verticalalignment", "bottom"))

        self.fig.canvas.draw()
        return self.fig.suptitle(label, y=y, **kwargs)

    @_defer
    def subplot_titles(self, *args, **kwargs):
        if args and isinstance(args[0], (list, tuple)):
            items = args[0]
            args = args[1:]
            return [
                subplot.title(item, *args, **kwargs)
                for item, subplot in zip(items, self.subplots)
            ]
        return [subplot.title(*args, **kwargs) for subplot in self.subplots]

    def show(self, *args, **kwargs):
        """Display the chart."""
        if len(self) == 0:
            self._rows, self._columns = (1, 1)
            self.add_subplot()
        self._release_queue()
        plt.show(*args, **kwargs)

    def save(self, *args, bbox_inches="tight", **kwargs):
        """Save the chart."""
        if len(self) == 0:
            self._rows, self._columns = (1, 1)
            self.add_subplot()
        self._release_queue()
        return plt.savefig(*args, bbox_inches=bbox_inches, **kwargs)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


_NO_CF_UNITS = False
try:
    import cf_units
except ImportError:
    _NO_CF_UNITS = True


TEMPERATURE_ANOM_UNITS = [
    "kelvin",
    "celsius",
]


PRETTY_UNITS = {
    "celsius": "°C",
    "fahrenheit": "°F",
}


def are_equal(unit_1, unit_2):
    if _NO_CF_UNITS:
        raise ImportError("cf-units is required for checking unit equivalence")
    return cf_units.Unit(unit_1) == cf_units.Unit(unit_2)


def anomaly_equivalence(units):
    return any(are_equal(units, t_units) for t_units in TEMPERATURE_ANOM_UNITS)


def convert(data, source_units, target_units):
    if _NO_CF_UNITS:
        raise ImportError("cf-units is required for unit conversion")
    return cf_units.Unit(source_units).convert(data, target_units)


def format_units(units):
    if _NO_CF_UNITS:
        return f"${PRETTY_UNITS.get(units, units)}$"

    from cf_units.tex import tex

    for name, formatted_units in PRETTY_UNITS.items():
        try:
            if are_equal(units, name):
                units = formatted_units
                break
        except ValueError:
            continue
    else:
        try:
            units = str(cf_units.Unit(units))
        except ValueError:
            pass

    try:
        formatted_units = f"${tex(units)}$"
    except SyntaxError:
        formatted_units = units

    return formatted_units



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import pytest

from earthkit.maps.domains import bounds


def test_from_bbox():
    assert bounds.from_bbox([-20, 40, -30, 50], ccrs.PlateCarree()) == [
        -20,
        40,
        -30,
        50,
    ]

    assert bounds.from_bbox(
        [-20, 40, -30, 50], ccrs.LambertAzimuthalEqualArea()
    ) == pytest.approx([-1985512, 3896680, -3476561, 5630595], 1)

    assert bounds.from_bbox(
        [-1985512, 3896680, -3476561, 5630595],
        ccrs.PlateCarree(),
        ccrs.LambertAzimuthalEqualArea(),
    ) == pytest.approx([-20, 40, -30, 50], 1)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from earthkit.maps.styles.magics import MagicsStyle

MAGICS_STYLE = {
    "contour": "off",
    "contour_description": "Method : Area fill Level range : -48 to 56 Interval : 4 Thickness : 3 Colour : All colours Used for temperature",
    "contour_hilo": "off",
    "contour_interval": "4",
    "contour_label": "off",
    "contour_legend_text": "Contour shade (Range: -48 / 56)",
    "contour_level_selection_type": "interval",
    "contour_line_contour": "dot",
    "contour_line_thickness": 3,
    "contour_shade": "on",
    "contour_shade_colour_list": "rgb(0,0,0.5)/rgb(0,0,0.5)/rgb(0,0,0.5)/rgb(0,0,0.5)/rgb(0,0,0.5)/rgb(0,0,0.85)/rgb(0.25,0,1)/blue_purple/greenish_blue/blue_green/bluish_green/yellow_green/greenish_yellow/yellow/orangish_yellow/orange_yellow/yellowish_orange/orange/reddish_orange/red_orange/orangish_red/red/magenta/magenta/magenta/magenta/magenta",
    "contour_shade_colour_method": "list",
    "contour_shade_max_level": 56,
    "contour_shade_method": "area_fill",
    "contour_shade_min_level": -48,
    "contour_title": "Contour shade (Range: -48 / 56)",
    "grib_missing_value_indicator": 9999,
}


def test_MagicsStyle_contour_levels_():
    style = MagicsStyle(**MAGICS_STYLE)
    assert style.levels.tolist() == list(range(-48, 56 + 4, 4))


def test_MagicsStyle_contour_colours():
    style = MagicsStyle(**MAGICS_STYLE)
    assert style.colors == [
        "#000080",
        "#000080",
        "#000080",
        "#000080",
        "#000080",
        "#0000d9",
        "#4000ff",
        "#8000ff",
        "#0080ff",
        "#00ffff",
        "#00ff80",
        "#80ff00",
        "#daff00",
        "#ffff00",
        "#fff500",
        "#ffda00",
        "#ffb000",
        "#ff8000",
        "#ff4f00",
        "#ff2500",
        "#ff0a00",
        "#ff0000",
        "#ff00ff",
        "#ff00ff",
        "#ff00ff",
        "#ff00ff",
        "#ff00ff",
    ]



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import pytest

from earthkit.maps.domains import natural_earth


def test_NaturalEarthDomain_France():
    domain = natural_earth.NaturalEarthDomain("France")

    assert domain.domain_name == "France"

    assert domain.geometry.__class__.__name__ == "MultiPolygon"

    assert isinstance(domain.crs, ccrs.AlbersEqualArea)

    assert isinstance(domain.bounds, list)


def test_NaturalEarthDomain_doesnt_exist():
    with pytest.raises(ValueError):
        natural_earth.NaturalEarthDomain("Tatooine").geometry



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from earthkit.maps.metadata import formatters, units
from earthkit.maps.metadata.formatters import TimeFormatter, format_month

__all__ = [
    "units",
    "formatters",
]


DEFAULT_FORECAST_TITLE = (
    "{variable_name}\n"
    "Base time: {base_time:%H:%M} on {base_time:%Y-%m-%d}   "
    "Valid time: {valid_time:%H:%M} on {valid_time:%Y-%m-%d} (T+{lead_time})"
)
DEFAULT_ANALYSIS_TITLE = (
    "{variable_name} at {valid_time:%H:%M} on {valid_time:%Y-%m-%d}"
)

TIME_KEYS = ["base_time", "valid_time", "lead_time", "time"]

MAGIC_KEYS = {
    "variable_name": {
        "preference": ["long_name", "standard_name", "name", "short_name"],
    },
    "short_name": {
        "preference": ["short_name", "name", "standard_name", "long_name"],
    },
    "month": {
        "function": format_month,
    },
}


def default_label(data):
    if data.metadata("type") == "an":
        format_string = DEFAULT_ANALYSIS_TITLE
    else:
        format_string = DEFAULT_FORECAST_TITLE
    return format_string


def extract(data, attr, default=None):

    if attr in TIME_KEYS:
        handler = TimeFormatter(data.datetime())
        label = getattr(handler, attr)[0]

    else:
        if hasattr(data, "metadata"):
            search = data.metadata
        else:
            data_key = [
                key
                for key in data.attrs["reduce_attrs"]
                if "reduce_dims" in data.attrs["reduce_attrs"][key]
            ][0]

            def search(x, default):
                return data.attrs["reduce_attrs"][data_key].get(x, default)

        candidates = [attr]
        if attr in MAGIC_KEYS:
            if "function" in MAGIC_KEYS[attr]:
                return MAGIC_KEYS[attr]["function"](data)
            else:
                candidates = MAGIC_KEYS[attr]["preference"] + candidates

        for item in candidates:
            label = search(item, default=None)
            if label is not None:
                break
        else:
            warnings.warn(f'No key "{attr}" found in layer metadata.')

    return label or default



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License"),
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast

import numpy as np
from matplotlib import colors

from earthkit.maps.styles.levels import Levels

MAGICS_COLOURS = {
    "automatic": (0.0, 0.0, 0.0),
    "none": (-1.0, -1.0, -1.0),
    "background": (1.0, 1.0, 1.0),
    "foreground": (0.0, 0.0, 0.0),
    "ecmwf_blue": (0.25, 0.43, 0.7),
    "red": (1.0000, 0.0000, 0.0000),
    "green": (0.0000, 1.0000, 0.0000),
    "blue": (0.0000, 0.0000, 1.0000),
    "yellow": (1.0000, 1.0000, 0.0000),
    "cyan": (0.0000, 1.0000, 1.0000),
    "magenta": (1.0000, 0.0000, 1.0000),
    "black": (0.0000, 0.0000, 0.0000),
    "avocado": (0.4225, 0.6500, 0.1950),
    "beige": (0.8500, 0.7178, 0.4675),
    "brick": (0.6000, 0.0844, 0.0300),
    "brown": (0.4078, 0.0643, 0.0000),
    "burgundy": (0.5000, 0.0000, 0.1727),
    "charcoal": (0.2000, 0.2000, 0.2000),
    "chestnut": (0.3200, 0.0112, 0.0000),
    "coral": (0.9000, 0.2895, 0.2250),
    "cream": (1.0000, 0.8860, 0.6700),
    "evergreen": (0.0000, 0.4500, 0.2945),
    "gold": (0.7500, 0.5751, 0.0750),
    "grey": (0.7000, 0.7000, 0.7000),
    "khaki": (0.5800, 0.4798, 0.2900),
    "kelly_green": (0.0000, 0.5500, 0.1900),
    "lavender": (0.6170, 0.4070, 0.9400),
    "mustard": (0.6000, 0.3927, 0.0000),
    "navy": (0.0000, 0.0000, 0.4000),
    "ochre": (0.6800, 0.4501, 0.0680),
    "olive": (0.3012, 0.3765, 0.0000),
    "peach": (0.9400, 0.4739, 0.3788),
    "pink": (0.9000, 0.3600, 0.4116),
    "rose": (0.8000, 0.2400, 0.4335),
    "rust": (0.7000, 0.2010, 0.0000),
    "sky": (0.4500, 0.6400, 1.0000),
    "tan": (0.4000, 0.3309, 0.2000),
    "tangerine": (0.8784, 0.4226, 0.0000),
    "turquoise": (0.1111, 0.7216, 0.6503),
    "violet": (0.4823, 0.0700, 0.7000),
    "reddish_purple": (1.0000, 0.0000, 0.8536),
    "purple_red": (1.0000, 0.0000, 0.5000),
    "purplish_red": (1.0000, 0.0000, 0.2730),
    "orangish_red": (1.0000, 0.0381, 0.0000),
    "red_orange": (1.0000, 0.1464, 0.0000),
    "reddish_orange": (1.0000, 0.3087, 0.0000),
    "orange": (1.0000, 0.5000, 0.0000),
    "yellowish_orange": (1.0000, 0.6913, 0.0000),
    "orange_yellow": (1.0000, 0.8536, 0.0000),
    "orangish_yellow": (1.0000, 0.9619, 0.0000),
    "greenish_yellow": (0.8536, 1.0000, 0.0000),
    "yellow_green": (0.5000, 1.0000, 0.0000),
    "yellowish_green": (0.1464, 1.0000, 0.0000),
    "bluish_green": (0.0000, 1.0000, 0.5000),
    "blue_green": (0.0000, 1.0000, 1.0000),
    "greenish_blue": (0.0000, 0.5000, 1.0000),
    "purplish_blue": (0.1464, 0.0000, 1.0000),
    "blue_purple": (0.5000, 0.0000, 1.0000),
    "bluish_purple": (0.8536, 0.0000, 1.0000),
    "purple": (1.0000, 0.0000, 1.0000),
    "white": (1.0000, 1.0000, 1.0000),
    "undefined": (-1.0, -1.0, -1.0),
}


class MagicsStyle:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def levels(self):
        selection_type = self.kwargs["contour_level_selection_type"]
        if selection_type == "list":
            return [float(i) for i in self.kwargs["contour_level_list"].split("/")]
        elif selection_type == "interval":
            interval = float(self.kwargs["contour_interval"])
            if "contour_shade_min_level" in self.kwargs:
                return np.arange(
                    float(self.kwargs["contour_shade_min_level"]),
                    float(self.kwargs["contour_shade_max_level"] + interval),
                    interval,
                )
            return Levels(
                step=interval,
                reference=self.kwargs.get("contour_reference_level"),
            )
        else:
            raise NotImplementedError(
                f"selection type '{selection_type}' is not yet supported"
            )

    @property
    def colors(self):
        colour_method = self.kwargs["contour_shade_colour_method"]
        if colour_method == "list":
            contour_colours = [
                parse_color(color)
                for color in self.kwargs["contour_shade_colour_list"].split("/")
            ]
        else:
            raise NotImplementedError(
                f"contour_shade_colour_method '{colour_method}' is not yet "
                f"supported"
            )
        return contour_colours


def parse_color(color):
    color = color.lower()
    if color.startswith("rgb"):
        color = ast.literal_eval(color.lstrip("rgb"))
    elif color.startswith("hsl"):
        color = colors.hsv_to_rgb(ast.literal_eval(color.lstrip("hsl")))
    else:
        color = MAGICS_COLOURS.get(color, color)
    if not isinstance(color, str):
        color = colors.rgb2hex(color)
    return color



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs

from earthkit.maps.domains import optimal


def test_Extents_area():
    extents = optimal.Extents([-50, 50, -30, 30])
    assert extents.area == 6000

    extents_wraps_globe = optimal.Extents([130, -130, -30, 30])
    assert extents_wraps_globe.area == 6000

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.area == 64800

    extents_null = optimal.Extents([-10, -10, 5, 5])
    assert extents_null.area == 0


def test_Extents_ratio():
    extents = optimal.Extents([-30, 30, -30, 30])
    assert extents.ratio == 1

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.ratio == 2

    extents_null = optimal.Extents([-10, -10, 5, 5])
    assert extents_null.ratio == 0


def test_Extents_standard_parallels():
    extents_northern_hemisphere = optimal.Extents([-20, 40, 30, 70])
    assert extents_northern_hemisphere.standard_parallels == (36.4, 63.6)

    extents_southern_hemisphere = optimal.Extents([-150, -70, -70, -30])
    assert extents_southern_hemisphere.standard_parallels == (-63.6, -36.4)


def test_Extents_central_lat():
    extents_northern_hemisphere = optimal.Extents([-20, 40, 30, 70])
    assert extents_northern_hemisphere.central_lat == 50

    extents_southern_hemisphere = optimal.Extents([-150, -70, -70, -30])
    assert extents_southern_hemisphere.central_lat == -50

    extents_equatorial = optimal.Extents([-150, -70, -30, 30])
    assert extents_equatorial.central_lat == 0

    extents_polar = optimal.Extents([-150, -70, 90, 90])
    assert extents_polar.central_lat == 90


def test_Extents_central_lon():
    extents_northern_hemisphere = optimal.Extents([-20, 40, 30, 70])
    assert extents_northern_hemisphere.central_lon == 10

    extents_southern_hemisphere = optimal.Extents([-150, -70, -70, -30])
    assert extents_southern_hemisphere.central_lon == -110

    extents_wraps_globe_positive = optimal.Extents([150, 230, -30, 30])
    assert extents_wraps_globe_positive.central_lon == 190

    extents_wraps_globe_switch = optimal.Extents([150, -130, -30, 30])
    assert extents_wraps_globe_switch.central_lon == 190


def test_Extents_is_landscape():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_landscape() is False

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_landscape() is True

    extents_tall = optimal.Extents([-30, 10, 10, 70])
    assert extents_tall.is_landscape() is False


def test_Extents_is_portrait():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_portrait() is False

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_portrait() is False

    extents_tall = optimal.Extents([-30, 10, 10, 70])
    assert extents_tall.is_portrait() is True


def test_Extents_is_square():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_square() is True

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_square() is False

    extents_tall = optimal.Extents([-30, 10, 10, 70])
    assert extents_tall.is_square() is False


def test_Extents_is_global():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_global() is False

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_global() is True

    extents_almost_global = optimal.Extents([-170, 170, -80, 80])
    assert extents_almost_global.is_global() is True

    extents_tall = optimal.Extents([-30, 10, 10, 70])
    assert extents_tall.is_global() is False


def test_Extents_is_large():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_large() is False

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_large() is False

    extents_large = optimal.Extents([-100, 90, -80, 80])
    assert extents_large.is_large() is True

    extents_tall = optimal.Extents([-30, 10, 10, 70])
    assert extents_tall.is_large() is False


def test_Extents_is_small():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_small() is True

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_small() is False

    extents_tall = optimal.Extents([-30, 10, 10, 70])
    assert extents_tall.is_small() is True


def test_Extents_is_polar():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_polar() is False

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_polar() is False

    extents_polar = optimal.Extents([-150, -70, 90, 90])
    assert extents_polar.is_polar() is True


def test_Extents_is_equatorial():
    extents_square = optimal.Extents([-30, 30, -30, 30])
    assert extents_square.is_equatorial() is True

    extents_global = optimal.Extents([-180, 180, -90, 90])
    assert extents_global.is_equatorial() is True

    extents_polar = optimal.Extents([-150, -70, 90, 90])
    assert extents_polar.is_equatorial() is False


def test_global():
    assert isinstance(
        optimal.crs_from_bounds([-180, 180, -90, 90]),
        ccrs.PlateCarree,
    )


def test_equatorial():
    assert isinstance(
        optimal.crs_from_bounds([-70, -20, -31, 29]),
        ccrs.PlateCarree,
    )


def test_north_polar():
    assert isinstance(
        optimal.crs_from_bounds([-180, 180, 70, 90]),
        ccrs.NorthPolarStereo,
    )


def test_south_polar():
    assert isinstance(
        optimal.crs_from_bounds([-180, 180, -90, -70]),
        ccrs.SouthPolarStereo,
    )


def test_landscape():
    assert isinstance(
        optimal.crs_from_bounds([-70, 30, 30, 70]),
        ccrs.AlbersEqualArea,
    )


def test_portrait():
    assert isinstance(
        optimal.crs_from_bounds([0, 30, 30, 70]),
        ccrs.TransverseMercator,
    )



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from string import Formatter

import numpy as np

from earthkit.maps import metadata, utils


class BaseFormatter(Formatter):
    """
    Formatter of earthkit-maps components, enabling convient titles and labels.
    """

    #: Attributes of subplots which can be extracted by format strings
    SUBPLOT_ATTRIBUTES = {
        "domain": "domain_name",
        "crs": "crs_name",
    }

    #: Attributes of styles which can be extracted by format strings
    STYLE_ATTRIBUTES = {
        "units": "units",
    }

    def convert_field(self, value, conversion):
        if conversion == "u":
            return str(value).upper()
        elif conversion == "l":
            return str(value).lower()
        return super().convert_field(value, conversion)

    def format_keys(self, format_string, kwargs):
        keys = (i[1] for i in self.parse(format_string) if i[1] is not None)
        for key in keys:
            kwargs[key] = self.format_key(key)
        return kwargs

    def format_key(self, key):
        return key

    def format(self, format_string, /, *args, **kwargs):
        kwargs = self.format_keys(format_string, kwargs)
        return super().format(format_string, *args, **kwargs)


class LayerFormatter(BaseFormatter):
    """
    Formatter of earthkit-maps `Layers`, enabling convient titles and labels.
    """

    def __init__(self, layer):
        self.layer = layer

    def format_keys(self, format_string, kwargs):
        keys = (i[1] for i in self.parse(format_string) if i[1] is not None)
        for key in keys:
            kwargs[key] = self.format_key(key)
        return kwargs

    def format_key(self, key):
        if key in self.SUBPLOT_ATTRIBUTES:
            value = getattr(self.layer.subplot, self.SUBPLOT_ATTRIBUTES[key])
        elif key in self.STYLE_ATTRIBUTES and self.layer.style is not None:
            value = getattr(self.layer.style, self.STYLE_ATTRIBUTES[key])
            if value is None:
                value = metadata.extract(self.layer.data, key)
                if key == "units":
                    value = metadata.units.format_units(value)
        else:
            value = metadata.extract(self.layer.data, key)
        return value


class SubplotFormatter(BaseFormatter):
    """
    Formatter of earthkit-maps `Subplots`, enabling convient titles and labels.
    """

    def __init__(self, subplot, unique=True):
        self.subplot = subplot
        self.unique = unique
        self._layer_index = None

    def convert_field(self, value, conversion):
        f = super().convert_field
        if isinstance(value, list):
            if isinstance(conversion, str) and conversion.isnumeric():
                return str(value[int(conversion)])
            return [f(v, conversion) for v in value]
        else:
            return f(value, conversion)

    def format_key(self, key):
        if key in self.SUBPLOT_ATTRIBUTES:
            values = [getattr(self.subplot, self.SUBPLOT_ATTRIBUTES[key])]
        else:
            values = [
                LayerFormatter(layer).format_key(key) for layer in self.subplot.layers
            ]
        return values

    def format_field(self, value, format_spec):
        f = super().format_field
        if isinstance(value, list):
            values = [f(v, format_spec) for v in value]
            if self._layer_index is not None:
                value = values[self._layer_index]
                self._layer_index = None
            else:
                if self.unique:
                    values = list(dict.fromkeys(values))
                value = utils.list_to_human(values)
        return value


class ChartFormatter(BaseFormatter):
    """
    Formatter of earthkit-maps `Charts`, enabling convient titles and labels.
    """

    def __init__(self, subplots, unique=True):
        self.subplots = subplots
        self.unique = unique
        self._layer_index = None

    def convert_field(self, value, conversion):
        f = super().convert_field
        if isinstance(value, list):
            if isinstance(conversion, str) and conversion.isnumeric():
                return str(value[int(conversion)])
            return [f(v, conversion) for v in value]
        else:
            return f(value, conversion)

    def format_key(self, key):
        values = [
            SubplotFormatter(subplot).format_key(key) for subplot in self.subplots
        ]
        values = [item for sublist in values for item in sublist]
        return values

    def format_field(self, value, format_spec):
        f = super().format_field
        if isinstance(value, list):
            values = [f(v, format_spec) for v in value]
            if self._layer_index is not None:
                value = values[self._layer_index]
                self._layer_index = None
            else:
                if self.unique:
                    values = list(dict.fromkeys(values))
                value = utils.list_to_human(values)
        return value


class TimeFormatter:
    def __init__(self, times):
        if not isinstance(times, (list, tuple)):
            times = [times]
        for i, time in enumerate(times):
            if not isinstance(time, dict):
                times[i] = {"time": time}
        self.times = times

    def _extract_time(method):
        def wrapper(self):
            attr = method.__name__
            times = [self._named_time(time, attr) for time in self.times]
            _, indices = np.unique(times, return_index=True)
            return [times[i] for i in sorted(indices)]

        return property(wrapper)

    @staticmethod
    def _named_time(time, attr):
        return time.get(attr, time.get("time"))

    @property
    def time(self):
        return self.valid_time

    @_extract_time
    def base_time(self):
        pass

    @_extract_time
    def valid_time(self):
        pass

    @property
    def lead_time(self):
        if len(self.base_time) == 1 and len(self.valid_time) > 1:
            times = itertools.product(self.base_time, self.valid_time)
        elif len(self.base_time) == len(self.valid_time):
            times = zip(self.base_time, self.valid_time)
        else:
            times = [
                (
                    self._named_time(time, "base_time"),
                    self._named_time(time, "valid_time"),
                )
                for time in self.times
            ]
        return [int((vtime - btime).total_seconds() / 3600) for btime, vtime in times]


def format_month(data):
    import calendar

    month = data.metadata("month", default=None)
    if month is not None:
        month = calendar.month_name[month]
    else:
        time = data.datetime()
        if "valid_time" in time:
            time = time["valid_time"]
        else:
            time = time["base_time"]
        month = f"{time:%B}"
    return month



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


COMPONENTS = {
    "u": {
        "criteria": [
            {"short_name": "10u"},
            {"short_name": "100u"},
            {"short_name": "u"},
            {"long_name": "10 metre U wind component"},
            {"variable": "u10"},
        ],
        "counterpart": "v",
    },
    "v": {
        "criteria": [
            {"short_name": "10v"},
            {"short_name": "100v"},
            {"short_name": "v"},
            {"long_name": "10 metre V wind component"},
            {"variable": "v10"},
        ],
        "counterpart": "u",
    },
}


def extract(data, components):
    results = []
    for component in components:
        for criteria in COMPONENTS[component]["criteria"]:
            for key, value in criteria.items():
                if value in data.metadata(key, default=None):
                    break
            else:
                continue
            break
        else:
            raise KeyError(f"No component '{component}' found in data")
        results.append(data.sel(**{key: value}))

    return results



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os

import yaml

from earthkit.maps import definitions, styles
from earthkit.maps.metadata.units import are_equal


def guess_style(data, units=None):
    from earthkit.maps import schema

    if os.path.exists(schema.style_library):
        styles_path = f"{schema.style_library}/*"
    else:
        styles_path = definitions.STYLES_DIR / (
            "*" if schema.style_library == "default" else f"{schema.style_library}/*"
        )

    for fname in glob.glob(str(styles_path)):
        if os.path.isfile(fname):
            with open(fname, "r") as f:
                config = yaml.load(f, Loader=yaml.SafeLoader)

        for criteria in config["criteria"]:
            for key, value in criteria.items():
                if data.metadata(key, default=None) != value:
                    break
            else:
                break
        else:
            continue
        break
    else:
        return styles.DEFAULT_STYLE

    if units is None:
        style = config["styles"][config["preferred-style"]]
    else:
        for _, style in config["styles"].items():
            if are_equal(style.get("units"), units):
                break
        else:
            # No style matching units found; return default
            return styles.Style(units=units)

    return styles.Style.from_dict(style)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


DEFAULT_LEGEND_LABEL = "{variable_name} ({units})"

_DISJOINT_LEGEND_LOCATIONS = {
    "bottom": {
        "loc": "upper center",
        "bbox_to_anchor": (0.5, -0.05),
    },
    "top": {
        "loc": "lower center",
        "bbox_to_anchor": (0.5, 1.0),
    },
    "left": {
        "loc": "upper right",
        "bbox_to_anchor": (-0.05, 1.0),
    },
    "right": {
        "loc": "upper left",
        "bbox_to_anchor": (1.05, 1.0),
    },
    "top left": {
        "loc": "lower center",
        "bbox_to_anchor": (0.25, 1),
    },
    "top right": {
        "loc": "lower center",
        "bbox_to_anchor": (0.75, 1),
    },
}


def colorbar(layer, *args, shrink=0.8, aspect=35, ax=None, **kwargs):
    """
    Produce a colorbar for a given layer.

    Parameters
    ----------
    layer : earthkit.maps.charts.layers.Layer
        The layer for which to produce a colorbar.
    **kwargs
        Any keyword arguments accepted by `matplotlib.figures.Figure.colorbar`.
    """
    label = kwargs.pop("label", DEFAULT_LEGEND_LABEL)
    label = layer.format_string(label)

    kwargs = {**layer.style._legend_kwargs, **kwargs}
    kwargs.setdefault("format", lambda x, _: f"{x:g}")

    if ax is None:
        kwargs["ax"] = layer.axes
    else:
        kwargs["cax"] = ax

    cbar = layer.fig.colorbar(
        layer.mappable,
        *args,
        label=label,
        shrink=shrink,
        aspect=aspect,
        **kwargs,
    )
    cbar.ax.minorticks_off()

    if cbar.solids is not None:
        cbar.solids.set(alpha=1)

    return cbar


def disjoint(layer, *args, location="bottom", frameon=False, **kwargs):
    """
    Produce a disjoint legend for a given layer.

    Parameters
    ----------
    layer : earthkit.maps.charts.layers.Layer
        The layer for which to produce a colorbar.
    **kwargs
        Any keyword arguments accepted by `matplotlib.figures.Figure.legend`.
    """
    kwargs.pop("format")  # remove higher-level kwargs which are invalid

    label = kwargs.pop("label", DEFAULT_LEGEND_LABEL)
    label = layer.format_string(label)

    source = layer.axes[0] if len(layer.axes) == 1 else layer.fig
    location_kwargs = _DISJOINT_LEGEND_LOCATIONS[location]

    artists, labels = layer.mappable.legend_elements()

    kwargs["ncols"] = kwargs.get("ncols", min(6, len(artists)))

    labels = kwargs.pop("labels", layer.style._bin_labels) or labels
    legend = source.legend(
        artists,
        labels,
        *args,
        title=label,
        frameon=frameon,
        **{**location_kwargs, **kwargs},
    )

    # Matplotlib removes legends when a new legend is plotted, so we have
    # to manually re-add them...
    if hasattr(layer.fig, "_previous_legend"):
        layer.fig.add_artist(layer.fig._previous_legend)
    layer.fig._previous_legend = legend

    layer.fig.canvas.draw()

    return legend



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import cartopy.io.shapereader as shpreader
import earthkit.data
import matplotlib.pyplot as plt
from adjustText import adjust_text

from earthkit.maps import domains, inputs, utils
from earthkit.maps.charts.layers import Layer
from earthkit.maps.domains import natural_earth
from earthkit.maps.metadata import components
from earthkit.maps.metadata.formatters import LayerFormatter, SubplotFormatter
from earthkit.maps.schemas import schema
from earthkit.maps.styles.levels import step_range


class Subplot:
    """
    An individual set of axes onto which one or more layer can be plotted.
    """

    @classmethod
    def from_data(cls, chart, data, *args, domain=None, crs=None, **kwargs):
        if not isinstance(data, earthkit.data.core.Base):
            data = earthkit.data.from_object(data)

        if domain is None and crs is None:
            try:
                crs = data.projection().to_cartopy_crs()
            except AttributeError:
                pass
        return cls(chart, *args, domain=domain, crs=crs, **kwargs)

    def __init__(self, chart, *args, domain=None, domain_crs=None, crs=None, **kwargs):
        if isinstance(domain, domains.Domain):
            self.domain = domain
            domain_crs = domain.crs
        else:
            self.domain = domains.parse(domain, crs)

        self.ax = chart.fig.add_subplot(*args, projection=self.domain.crs, **kwargs)

        if self.domain.bounds is not None:
            self.ax.set_extent(self.domain.bounds, crs=domain_crs or self.domain.crs)

        self.chart = chart
        self.layers = []

    @property
    def fig(self):
        return self.chart.fig

    @property
    def crs_name(self):
        return self.domain.crs.__class__.__name__

    @property
    def domain_name(self):
        name = self.domain.title
        if name == "None":
            if self.domain.crs.__class__.__name__ == "PlateCarree":
                extent = self.ax.get_extent()
                if extent == (-180.0, 180.0, -90.0, 90.0):
                    name = "Global"
                else:
                    name = domains.bounds.to_string(extent)
        return name

    @property
    def distinct_legend_layers(self):
        """Layers on this subplot which have a unique `Style`."""
        unique_layers = []
        for layer in self.layers:
            for legend_layer in unique_layers:
                if legend_layer.style == layer.style:
                    break
            else:
                unique_layers.append(layer)
        return unique_layers

    def _can_transform_first(self, method):
        """
        Check if the data can be transformed onto the target projection.

        Some matplotlib methods can be sped up by transforming data coordinates
        before plotting, but this is not compatible with all projections nor all
        plotting methods. This method checks if a given domain and method is
        compatible with `transform_first`.

        Parameters
        ----------
        method : method
            The class method to check for compatibility with `transform_first`.

        Returns
        -------
        bool
        """
        if self.domain._can_transform_first and method.__name__ != "pcolormesh":
            return True
        else:
            return False

    def plot_gridded_scalar(method):
        """
        Decorator for transforming input data into plottable components.

        Parameters
        ----------
        method : method
            The method for which to transform input parameters to plottable
            components.
        """

        def wrapper(self, data, *args, transform_first=None, style=None, **kwargs):
            data = inputs.sanitise(data)

            input_data = inputs.Input(
                data, *args, domain=self.domain, style=style, **kwargs
            )

            kwargs.pop("x", None)
            kwargs.pop("y", None)
            kwargs.pop("units", None)

            if transform_first is None:
                transform_first = self._can_transform_first(method)

            mappable = method(
                self,
                input_data.x,
                input_data.y,
                input_data.values,
                style=input_data.style,
                transform=input_data.transform,
                transform_first=transform_first,
                **kwargs,
            )

            layer = Layer(data, mappable, self, style=input_data.style)
            self.layers.append(layer)
            return layer

        return wrapper

    def plot_gridded_vector(method):
        """
        Decorator for transforming input data into plottable components.

        Parameters
        ----------
        method : method
            The method for which to transform input parameters to plottable
            components.
        """

        def wrapper(self, data, transform_first=None, style=None, **kwargs):
            if len(data) == 2:
                u, v = data
            else:
                u, v = components.extract(*data, "uv")

            u = inputs.sanitise(u)
            v = inputs.sanitise(v)

            input_data = inputs.Vector(u, v, domain=self.domain, style=style, **kwargs)

            kwargs.pop("x", None)
            kwargs.pop("y", None)
            kwargs.pop("units", None)

            if transform_first is None:
                transform_first = self._can_transform_first(method)

            mappable = method(
                self,
                input_data.x,
                input_data.y,
                input_data.u,
                input_data.v,
                style=input_data.style,
                transform=input_data.transform,
                **kwargs,
            )

            layer = Layer([u, v], mappable, self, style=input_data.style)
            self.layers.append(layer)
            return layer

        return wrapper

    def plot(self, data, *args, **kwargs):
        if data.__class__.__name__ in ("DataFrame", "Series"):
            raise NotImplementedError(
                "Plotting pandas DataFrames and Series is an upcoming feature"
            )
        else:
            return self._plot_gridded_scalar(data, *args, **kwargs)

    @plot_gridded_scalar
    def _plot_gridded_scalar(self, *args, style=None, **kwargs):
        try:
            return style.plot(self.ax, *args, **kwargs)
        except NotImplementedError:
            return style.contourf(self.ax, *args, **kwargs)

    @plot_gridded_scalar
    def contourf(self, *args, style=None, **kwargs):
        return style.contourf(self.ax, *args, **kwargs)

    shaded_contour = contourf  # alias to contourf

    @plot_gridded_scalar
    def contour(self, *args, style=None, **kwargs):
        return style.contour(self.ax, *args, **kwargs)

    @plot_gridded_scalar
    def pcolormesh(self, *args, style=None, **kwargs):
        return style.pcolormesh(self.ax, *args, **kwargs)

    @plot_gridded_scalar
    def scatter(self, *args, style=None, **kwargs):
        return style.scatter(self.ax, *args, **kwargs)

    @plot_gridded_vector
    @schema.barbs.apply()
    def barbs(self, *args, style=None, **kwargs):
        style.barbs(self.ax, *args, **kwargs)

    @schema.coastlines.apply()
    def coastlines(self, *args, resolution="auto", **kwargs):
        """Add coastal outlines from the Natural Earth “coastline” collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        return self.ax.coastlines(*args, resolution=resolution, **kwargs)

    @schema.borders.apply()
    def borders(
        self, *args, resolution="auto", labels=False, label_kwargs=None, **kwargs
    ):
        """Add country boundary polygons from Natural Earth.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.BORDERS
        else:
            feature = cfeature.NaturalEarthFeature(
                "cultural", "admin_0_countries", resolution
            )

        if labels:
            resolution = "110m" if resolution == "auto" else resolution
            shpfilename = shpreader.natural_earth(
                resolution=resolution, category="cultural", name="admin_0_countries"
            )
            reader = shpreader.Reader(shpfilename)
            if not isinstance(labels, str):
                labels = "ISO_A2_EH"
            self._add_polygon_labels(
                reader, x_key="LABEL_X", y_key="LABEL_Y", label_key=labels
            )

        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.shapes.apply()
    def shapes(
        self,
        shapes,
        *args,
        transform=ccrs.PlateCarree(),
        adjust_labels=False,
        labels=False,
        **kwargs,
    ):
        if isinstance(shapes, str):
            shapes = shpreader.Reader(shapes)
        results = self.ax.add_geometries(
            shapes.geometries(), transform, *args, **kwargs
        )
        if labels:
            label_key = labels if isinstance(labels, str) else None
            self._add_polygon_labels(
                shapes, label_key=label_key, adjust_labels=adjust_labels
            )
        return results

    def _add_polygon_labels(
        self, reader, x_key=None, y_key=None, adjust_labels=False, label_key=None
    ):
        records = list(reader.records())
        label_kwargs = dict()
        label_kwargs = {
            **dict(
                ha="center",
                va="center",
                bbox=dict(
                    boxstyle="round",
                    ec=(0.2, 0.2, 0.2, 0),
                    fc=(0.3, 0.3, 0.3),
                ),
                fontsize=8,
                weight="bold",
                color=(0.95, 0.95, 0.95),
                clip_on=True,
                clip_box=self.ax.bbox,
                transform=ccrs.Geodetic(),
            ),
            **label_kwargs,
        }

        if label_key is None:
            for label_key in records[0].attributes:
                if "name" in label_key.lower():
                    break

        texts = []
        for record in reader.records():
            name = record.attributes[label_key]

            if record.geometry.__class__.__name__ == "MultiPolygon":
                centroid = max(record.geometry.geoms, key=lambda a: a.area).centroid
            else:
                centroid = record.geometry.centroid
            x = centroid.x
            y = centroid.y
            if x_key:
                x = record.attributes[x_key]
            if y_key:
                y = record.attributes[y_key]

            text = self.ax.text(x, y, name, **label_kwargs)
            texts.append(text)
        if adjust_labels:
            adjust_text(texts)
        return texts

    @schema.states_provinces.apply()
    def states_provinces(
        self, *args, resolution="auto", labels=False, adjust_labels=False, **kwargs
    ):
        """Add country boundary polygons from Natural Earth.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.STATES
        else:
            feature = cfeature.NaturalEarthFeature(
                "cultural",
                "admin_1_states_provinces",
                resolution,
                facecolor="none",
            )

        if labels:
            resolution = "110m" if resolution == "auto" else resolution
            shpfilename = shpreader.natural_earth(
                resolution=resolution,
                category="cultural",
                name="admin_1_states_provinces",
            )
            reader = shpreader.Reader(shpfilename)
            if not isinstance(labels, str):
                labels = "name"
            self._add_polygon_labels(
                reader, label_key=labels, adjust_labels=adjust_labels
            )

        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.land.apply()
    def land(self, *args, resolution="auto", **kwargs):
        """Add land polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.LAND
        else:
            feature = cfeature.NaturalEarthFeature("physical", "land", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.ocean.apply()
    def ocean(self, *args, resolution="auto", **kwargs):
        """Add ocean polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.OCEAN
        else:
            feature = cfeature.NaturalEarthFeature("physical", "ocean", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.rivers.apply()
    def rivers(self, *args, resolution="auto", **kwargs):
        """Add ocean polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.RIVERS
        else:
            feature = cfeature.NaturalEarthFeature(
                "physical",
                "rivers_lake_centerlines",
                resolution,
                facecolor="none",
                edgecolor=kwargs.pop("edgecolor", "#C3EDFF"),
            )
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.urban_areas.apply()
    def urban_areas(self, *args, resolution="50m", **kwargs):
        """Add urban area polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        feature = cfeature.NaturalEarthFeature("cultural", "urban_areas", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    @schema.lakes.apply()
    def lakes(self, *args, resolution="50m", **kwargs):
        """Add lake polygons from the Natural Earth collection.

        Parameters
        ----------
        resolution: (str, optional)
            One of "low", "medium" or "high", or a named resolution from the
            Natrual Earth dataset.
        """
        resolution = natural_earth.RESOLUTIONS.get(resolution, resolution)
        if resolution == "auto":
            feature = cfeature.LAKES
        else:
            feature = cfeature.NaturalEarthFeature("physical", "lakes", resolution)
        return self.ax.add_feature(feature, *args, **kwargs)

    def cities(
        self,
        *args,
        density="medium",
        color="black",
        capitals_only=False,
        adjust_labels=True,
        **kwargs,
    ):
        density = natural_earth.RESOLUTIONS.get(density, density)
        fname = shpreader.natural_earth(
            resolution=density,
            category="cultural",
            name="populated_places",
        )
        reader = shpreader.Reader(fname)
        records = list(reader.records())

        texts = []
        for record in records:
            if capitals_only and not record.attributes["ADM0CAP"]:
                continue
            if self.domain.contains_point(
                (record.geometry.x, record.geometry.y),
                crs=ccrs.PlateCarree(),
            ):
                scatter_kwargs = {
                    "marker": "o",
                    "s": 10,
                    "color": "red",
                    "edgecolor": "black",
                    "linewidths": 0.5,
                }
                text_kwargs = {
                    "fontsize": 7,
                    "color": color,
                }
                if record.attributes["ADM0CAP"]:
                    scatter_kwargs = {
                        "marker": "s",
                        "s": 25,
                        "color": "goldenrod",
                        "edgecolor": "black",
                        "linewidths": 0.5,
                    }
                    text_kwargs = {
                        "fontsize": 8,
                        "fontweight": "bold",
                        "color": color,
                    }
                elif record.attributes["RANK_MAX"] < 8:
                    scatter_kwargs = {
                        "marker": "o",
                        "s": 6,
                        "color": "navy",
                    }
                    text_kwargs = {
                        "fontsize": 7,
                        "fontstyle": "italic",
                        "color": color,
                    }
                self.ax.scatter(
                    record.geometry.x,
                    record.geometry.y,
                    transform=ccrs.PlateCarree(),
                    zorder=10,
                    **scatter_kwargs,
                )
                text = self.ax.text(
                    record.geometry.x,
                    record.geometry.y,
                    record.attributes["NAME_EN"],
                    transform=ccrs.PlateCarree(),
                    clip_on=True,
                    zorder=10,
                    **text_kwargs,
                )
                texts.append(text)
        if adjust_labels:
            warnings.warn(
                "City label positions are being automatically adjusted to "
                "minimise overlaps, which can take a long time. This behaviour "
                "can be switched off by passing `adjust_labels=False`."
            )
            adjust_text(texts)
        return texts

    def image(self, img, extent, origin="upper", transform=None):
        if isinstance(img, str):
            import PIL

            img = PIL.Image.open(img)
        if transform is None:
            transform = ccrs.PlateCarree()
        return self.ax.imshow(img, origin=origin, extent=extent, transform=transform)

    def stock_img(self, *args, **kwargs):
        self.ax.stock_img(*args, **kwargs)

    @schema.gridlines.apply()
    def gridlines(self, *args, xstep=None, xref=0, ystep=None, yref=0, **kwargs):
        """
        Add gridlines to the subplot.

        Parameters
        ----------
        xstep : float, optional
            The step/difference between each x gridline.
        xref : float, optional
            The reference point around which to calibrate the x level range.
        ystep : float, optional
            The step/difference between each y gridline.
        yref : float, optional
            The reference point around which to calibrate the y level range.

        Note
        ----
            Any keyword arguments accepted by cartopy's `gridlines` method can
            be used when adding gridlines.
        """
        if xstep is not None:
            kwargs["xlocs"] = step_range([-180, 180], xstep, xref)
        if ystep is not None:
            kwargs["ylocs"] = step_range([-90, 90], ystep, yref)
        self._gridlines = self.ax.gridlines(*args, **kwargs)
        return self._gridlines

    @schema.legend.apply()
    def legend(self, *args, **kwargs):
        legends = []
        for layer in self.distinct_legend_layers:
            if layer.style is not None:
                legend = layer.style.legend(layer, *args, **kwargs)
            legends.append(legend)
        return legends

    @property
    def _default_title_template(self):
        templates = [layer._default_title_template for layer in self.layers]
        if len(set(templates)) == 1:
            template = templates[0]
        else:
            title_parts = []
            for i, template in enumerate(templates):
                keys = [k for _, k, _, _ in SubplotFormatter().parse(template)]
                for key in set(keys):
                    template = template.replace("{" + key, "{" + key + f"!{i}")
                title_parts.append(template)
            template = utils.list_to_human(title_parts)
        return template

    @schema.title.apply()
    def title(self, label=None, unique=True, wrap=True, **kwargs):
        if label is None:
            label = self._default_title_template
        label = self.format_string(label, unique)
        plt.sca(self.ax)
        return plt.title(label, wrap=wrap, **kwargs)

    def format_string(self, string, unique=True, grouped=True):
        if not grouped:
            return utils.list_to_human(
                [LayerFormatter(layer).format(string) for layer in self.layers]
            )
        else:
            return SubplotFormatter(self, unique=unique).format(string)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections


def recursive_dict_update(original_dict, update_dict):
    """
    Recursively update a dictionary with keys and values from another dictionary.

    Parameters
    ----------
    original_dict : dict
        The original dictionary to be updated (in place).
    update_dict : dict
        The dictionary containing keys to be updated in the original dictionary.
    """
    for k, v in update_dict.items():
        if isinstance(v, collections.abc.Mapping):
            original_dict[k] = recursive_dict_update(original_dict.get(k, {}), v)
        else:
            original_dict[k] = v
    return original_dict


def list_to_human(iterable, conjunction="and", oxford_comma=False):
    """
    Convert an iterable to a human-readable string.

    Parameters
    ----------
    iterable : list or tuple
        The list of strings to convert to a single natural language string.
    conjunction : str, optional
        The conjunction with which to join the last two elements of the list,
        for example "and" (default).
    oxford_comma : bool, optional
        If `True`, an "Oxford comma" will be added before the conjunction when
        there are three or more elements in the list. Default is `False`.
    """
    list_of_strs = [str(item) for item in iterable]

    if len(list_of_strs) > 2:
        list_of_strs = [", ".join(list_of_strs[:-1]), list_of_strs[-1]]
        if oxford_comma:
            list_of_strs[0] += ","

    return f" {conjunction} ".join(list_of_strs)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib as mpl
import numpy as np
from matplotlib.colors import BoundaryNorm, LinearSegmentedColormap, ListedColormap


def expand(colors, levels, extend_colors=0):
    """
    Generate a list of colours from a matplotlib colormap name and some levels.
    """
    length = len(levels) + extend_colors

    if isinstance(colors, (list, tuple)) and len(colors) == 1:
        colors *= length - 1
    if isinstance(colors, str):
        try:
            cmap = mpl.colormaps[colors]
        except KeyError:
            colors = [colors] * (length - 1)
        else:
            colors = [cmap(i) for i in np.linspace(0, 1, length)]
    return colors


def contour_line_colors(colors, levels):
    colors = expand(colors, levels)
    cmap = LinearSegmentedColormap.from_list(name="", colors=colors, N=len(levels))
    return cmap


def cmap_and_norm(colors, levels, normalize=True, extend=None):

    levels = list(levels)
    extend_colors = 0
    if extend == "both":
        extend_colors = 2
        levels = [-np.inf] + levels + [np.inf]
    elif extend == "max":
        levels += [np.inf]
        extend_colors = 1
    elif extend == "min":
        levels = [-np.inf] + levels
        extend_colors = 1

    colors = expand(colors, levels, extend_colors)
    N = len(levels) + extend_colors - 1

    colormap = LinearSegmentedColormap.from_list
    if len(colors) == N:
        colormap = ListedColormap

    cmap = colormap(name="", colors=colors, N=N)

    norm = None
    if normalize:
        norm = BoundaryNorm(levels, cmap.N)

    return cmap, norm


def gradients(levels, colors, gradients, normalize, **kwargs):

    normalised = (levels - np.min(levels)) / (np.max(levels) - np.min(levels))
    color_bins = list(zip(normalised, colors))
    cmap = LinearSegmentedColormap.from_list(name="", colors=color_bins, N=255)

    if not isinstance(gradients, (list, tuple)):
        gradients = [gradients] * (len(levels) - 1)

    extrapolated_levels = []
    for i in range(len(levels) - 1):
        bins = list(np.linspace(levels[i], levels[i + 1], gradients[i]))
        extrapolated_levels += bins[(1 if i != 0 else 0) :]
    levels = extrapolated_levels

    norm = None
    if normalize:
        norm = BoundaryNorm(levels, cmap.N)

    return {**{"cmap": cmap, "norm": norm, "levels": levels}, **kwargs}



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from earthkit.maps.styles import levels


def test_auto_range():
    # diverging at zero
    assert levels.auto_range(
        [-1, 4],
        divergence_point=0,
        n_levels=5,
    ) == [-4, -2, 0, 2, 4]


def test_step_range():
    assert levels.step_range([3], step=4) == [0, 4]
    assert levels.step_range([2], step=4, reference=3) == [-1, 3]

    assert levels.step_range([3.2], step=0.5) == [3, 3.5]

    assert levels.step_range([1, 3, 7], step=4) == [0, 4, 8]
    assert levels.step_range([1, 3, 7], step=4, reference=2) == [-2, 2, 6, 10]

    assert levels.step_range(
        [983, 1023],
        step=4,
        reference=1000,
    ) == [980, 984, 988, 992, 996, 1000, 1004, 1008, 1012, 1016, 1020, 1024]

    assert levels.step_range(
        [983, 1023],
        step=4,
        reference=1001,
    ) == [981, 985, 989, 993, 997, 1001, 1005, 1009, 1013, 1017, 1021, 1025]


def test_Levels():
    # diverging at zero
    assert levels.Levels(divergence_point=0).apply(
        [-5, 8],
    ) == [-8, -6, -4, -2, 0, 2, 4, 6, 8]

    assert levels.Levels(step=4).apply([3]) == [0, 4]
    assert levels.Levels(step=4, reference=3).apply([2]) == [-1, 3]

    assert levels.Levels(step=4, reference=1000).apply(
        [983, 1023],
    ) == [980, 984, 988, 992, 996, 1000, 1004, 1008, 1012, 1016, 1020, 1024]



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from earthkit.maps.styles import colors


def test_expand():
    assert colors.expand(["red"], [1, 2]) == ["red"]

    assert colors.expand("viridis", [1, 2])[0] == pytest.approx(
        [0.267, 0.005, 0.329, 1.000], 0.1
    )

    assert colors.expand("viridis", [1, 2])[1] == pytest.approx(
        [0.993, 0.906, 0.144, 1.000], 0.1
    )



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs
import pytest
from earthkit.data.utils.bbox import BoundingBox

from earthkit.maps.domains import parse


def test_parse_string():
    domain = parse("France", None)

    assert domain.domain_name == "France"
    assert domain.bounds == pytest.approx([-626161, 690812, -625183, 649633], 1)
    assert isinstance(domain.crs, ccrs.AlbersEqualArea)


def test_parse_None():
    domain = parse(None, None)

    assert domain.domain_name is None
    assert isinstance(domain.crs, ccrs.PlateCarree)


def test_parse_extents():
    domain = parse([-20, 40, 30, 72], None)
    assert isinstance(domain.crs, ccrs.AlbersEqualArea)


def test_parse_crs():
    domain = parse(None, crs=ccrs.EckertIV())
    assert isinstance(domain.crs, ccrs.EckertIV)


def test_parse_bbox():
    domain = parse(BoundingBox(north=72, south=30, east=-20, west=40), None)
    assert domain.bounds == pytest.approx([-6932732, 6932732, -2352422, 6281727], 1)
    assert isinstance(domain.crs, ccrs.AlbersEqualArea)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from earthkit.maps import styles
from earthkit.maps.metadata import units


def test_Style_static_levels():
    style = styles.Style(levels=[1, 2, 3])
    assert style.levels([1]) == [1, 2, 3]
    assert style.levels([-19999]) == [1, 2, 3]
    assert style.levels(range(9999)) == [1, 2, 3]


def test_Style_levels():
    static_style = styles.Style(levels=[1, 2, 3])
    assert static_style.levels() == [1, 2, 3]

    range_style = styles.Style(levels=range(-10, 11, 5))
    assert range_style.levels() == range(-10, 11, 5)

    dynamic_style = styles.Style()
    with pytest.raises(ValueError):
        dynamic_style.levels()


@pytest.mark.skipif(units._NO_CF_UNITS, reason="cf-units is not installed")
def test_Style_units():
    style = styles.Style(units="celsius")
    assert style.units == "$°C$"


@pytest.mark.skipif(units._NO_CF_UNITS, reason="cf-units is not installed")
def test_Style_convert_units():
    style = styles.Style(units="celsius")
    assert style.convert_units(273.15, source_units="kelvin") == 0


@pytest.mark.skipif(units._NO_CF_UNITS, reason="cf-units is not installed")
def test_Style_convert_units_anomaly():
    style = styles.Style(units="celsius")
    assert (
        style.convert_units(-5, source_units="kelvin", short_name="temperature_anomaly")
        == -5
    )



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs

from earthkit.maps.domains import domain


def test_Domain_from_string():
    greenland = domain.Domain.from_string("greenland")
    assert greenland.title == "Greenland"
    assert isinstance(greenland.crs, ccrs.AlbersEqualArea)

    europe = domain.Domain.from_string("europe")
    assert europe.title == "Europe"
    assert isinstance(europe.crs, ccrs.LambertAzimuthalEqualArea)



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from earthkit.maps.metadata import formatters


def test_BaseFormatter_conversion():
    assert formatters.BaseFormatter().format("My {test!u}") == "My TEST"
    assert formatters.BaseFormatter().format("My {TeSt!l}") == "My test"


def test_TimeFormatter_time():
    time = datetime(2020, 1, 1)
    assert formatters.TimeFormatter(time).time == [datetime(2020, 1, 1)]


def test_TimeFormatter_base_time():
    time = {
        "base_time": datetime(2020, 1, 1),
        "valid_time": datetime(2020, 1, 7),
    }
    assert formatters.TimeFormatter(time).base_time == [datetime(2020, 1, 1)]


def test_TimeFormatter_valid_time():
    time = {
        "base_time": datetime(2020, 1, 1),
        "valid_time": datetime(2020, 1, 7),
    }
    assert formatters.TimeFormatter(time).valid_time == [datetime(2020, 1, 7)]


def test_TimeFormatter_lead_time():
    time = {
        "base_time": datetime(2020, 1, 1),
        "valid_time": datetime(2020, 1, 7),
    }
    assert formatters.TimeFormatter(time).lead_time == [144]



# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from earthkit.maps import metadata, utils
from earthkit.maps.metadata.formatters import LayerFormatter


class Layer:
    """
    Class connecting a plotted map layer with its underlying data.

    Parameters
    ----------
    data : earthkit.data.core.Base
        The underlying data represented on the plot.
    """

    def __init__(self, data, mappable, subplot, style=None):
        self.data = data
        self.mappable = mappable
        self.subplot = subplot
        self.style = style

    @property
    def fig(self):
        """The matplotlib figure on which this layer is plotted."""
        return self.subplot.fig

    @property
    def ax(self):
        """The matplotlib axes on which this layer is plotted."""
        return self.subplot.ax

    @property
    def axes(self):
        """All matplotlib axes over which this layer is plotted."""
        return [self.ax]

    def legend(self, *args, **kwargs):
        """
        Generate a legend for this specific layer.
        """
        if self.style is not None:
            return self.style.legend(self, *args, **kwargs)

    def format_string(self, string):
        return LayerFormatter(self).format(string)

    @property
    def _default_title_template(self):
        if self.data.metadata("type", default="an") == "an":
            template = metadata.DEFAULT_ANALYSIS_TITLE
        else:
            template = metadata.DEFAULT_FORECAST_TITLE
        return template


class LayerGroup:
    """
    A group of related layers.

    Parameters
    ----------
    layers : list of earthkit.maps.charts.layers.Layer objects
        A list of grouped layers.
    """

    def __init__(self, layers):
        self.layers = layers

    @property
    def subplots(self):
        return [layer.subplot for layer in self.layers]

    @property
    def fig(self):
        return self.subplots[0].fig

    @property
    def axes(self):
        return [subplot.ax for subplot in self.subplots]

    @property
    def style(self):
        return self.layers[0].style

    def legend(self, *args, **kwargs):
        if self.style is not None:
            return self.style.legend(
                self.fig,
                self,
                *args,
                ax=self.axes,
                **kwargs,
            )

    @property
    def mappable(self):
        return self.layers[0].mappable

    def format_string(self, string, unique=True):
        results = [layer.format_string(string) for layer in self.layers]

        if unique:
            results = list(dict.fromkeys(results))

        result = utils.list_to_human(results)
        return result
